"""pago20 http://www.sat.gob.mx/Pagos20"""
from collections.abc import *
from datetime import datetime
from decimal import Decimal

from ..compute import make_impuestos_p, make_pago_totales
from ... import iterate
from ... import CFDI, XElement, ScalarMap


class TrasladoDR(ScalarMap):
    """
    Nodo requerido para asentar la información detallada de un traslado de impuesto específico conforme al monto del pago recibido.

    :param base_dr: Atributo requerido para señalar la base para el cálculo del impuesto trasladado conforme al monto del pago, aplicable al documento relacionado, la determinación de la base se realiza de acuerdo con las disposiciones fiscales vigentes. No se permiten valores negativos.
    :param impuesto_dr: Atributo requerido para señalar la clave del tipo de impuesto trasladado conforme al monto del pago, aplicable al documento relacionado.
    :param tipo_factor_dr: Atributo requerido para señalar la clave del tipo de factor que se aplica a la base del impuesto.
    :param tasa_o_cuota_dr: Atributo condicional para señalar el valor de la tasa o cuota del impuesto que se traslada. Es requerido cuando el atributo TipoFactorDR contenga una clave que corresponda a Tasa o Cuota.
    :param importe_dr: Atributo condicional para señalar el importe del impuesto trasladado conforme al monto del pago, aplicable al documento relacionado. No se permiten valores negativos. Es requerido cuando el tipo factor sea Tasa o Cuota.
    """

    def __init__(
            self,
            base_dr: Decimal | int,
            impuesto_dr: str,
            tipo_factor_dr: str,
            tasa_o_cuota_dr: Decimal | int = None,
            importe_dr: Decimal | int = None,
    ):
        super().__init__({
            'BaseDR': base_dr,
            'ImpuestoDR': impuesto_dr,
            'TipoFactorDR': tipo_factor_dr,
            'TasaOCuotaDR': tasa_o_cuota_dr,
            'ImporteDR': importe_dr,
        })
        

class RetencionDR(ScalarMap):
    """
    Nodo requerido para registrar la información detallada de una retención de impuesto específico conforme al monto del pago recibido.

    :param base_dr: Atributo requerido para señalar la base para el cálculo de la retención conforme al monto del pago, aplicable al documento relacionado, la determinación de la base se realiza de acuerdo con las disposiciones fiscales vigentes. No se permiten valores negativos.
    :param impuesto_dr: Atributo requerido para señalar la clave del tipo de impuesto retenido conforme al monto del pago, aplicable al documento relacionado.
    :param tipo_factor_dr: Atributo requerido para señalar la clave del tipo de factor que se aplica a la base del impuesto.
    :param tasa_o_cuota_dr: Atributo requerido para señalar el valor de la tasa o cuota del impuesto que se retiene.
    :param importe_dr: Atributo requerido para señalar el importe del impuesto retenido conforme al monto del pago, aplicable al documento relacionado. No se permiten valores negativos.
    """

    def __init__(
            self,
            base_dr: Decimal | int,
            impuesto_dr: str,
            tipo_factor_dr: str,
            tasa_o_cuota_dr: Decimal | int,
            importe_dr: Decimal | int,
    ):
        super().__init__({
            'BaseDR': base_dr,
            'ImpuestoDR': impuesto_dr,
            'TipoFactorDR': tipo_factor_dr,
            'TasaOCuotaDR': tasa_o_cuota_dr,
            'ImporteDR': importe_dr,
        })
        

class ImpuestosDR(ScalarMap):
    """
    Nodo condicional para registrar los impuestos aplicables conforme al monto del pago recibido, expresados a la moneda del documento relacionado.

    :param retenciones_dr: Nodo opcional para capturar los impuestos retenidos aplicables conforme al monto del pago recibido.
    :param traslados_dr: Nodo opcional para capturar los impuestos trasladados aplicables conforme al monto del pago recibido.
    """

    def __init__(
            self,
            retenciones_dr: Sequence[RetencionDR | dict] = None,
            traslados_dr: Sequence[TrasladoDR | dict] = None,
    ):
        super().__init__({
            'RetencionesDR': retenciones_dr,
            'TrasladosDR': traslados_dr,
        })
        

class DoctoRelacionado(ScalarMap):
    """
    Nodo requerido para expresar la lista de documentos relacionados con los pagos. Por cada documento que se relacione se debe generar un nodo DoctoRelacionado.

    :param id_documento: Atributo requerido para expresar el identificador del documento relacionado con el pago. Este dato puede ser un Folio Fiscal de la Factura Electrónica o bien el número de operación de un documento digital.
    :param moneda_dr: Atributo requerido para identificar la clave de la moneda utilizada en los importes del documento relacionado, cuando se usa moneda nacional o el documento relacionado no especifica la moneda se registra MXN. Los importes registrados en los atributos “ImpSaldoAnt”, “ImpPagado” e “ImpSaldoInsoluto” de éste nodo, deben corresponder a esta moneda. Conforme con la especificación ISO 4217.
    :param num_parcialidad: Atributo requerido para expresar el número de parcialidad que corresponde al pago.
    :param imp_saldo_ant: Atributo requerido para expresar el monto del saldo insoluto de la parcialidad anterior. En el caso de que sea la primer parcialidad este atributo debe contener el importe total del documento relacionado.
    :param imp_pagado: Atributo requerido para expresar el importe pagado para el documento relacionado.
    :param objeto_imp_dr: Atributo requerido para expresar si el pago del documento relacionado es objeto o no de impuesto.
    :param serie: Atributo opcional para precisar la serie del comprobante para control interno del contribuyente, acepta una cadena de caracteres.
    :param folio: Atributo opcional para precisar el folio del comprobante para control interno del contribuyente, acepta una cadena de caracteres.
    :param equivalencia_dr: Atributo condicional para expresar el tipo de cambio conforme con la moneda registrada en el documento relacionado. Es requerido cuando la moneda del documento relacionado es distinta de la moneda de pago. Se debe registrar el número de unidades de la moneda señalada en el documento relacionado que equivalen a una unidad de la moneda del pago. Por ejemplo: El documento relacionado se registra en USD. El pago se realiza por 100 EUR. Este atributo se registra como 1.114700 USD/EUR. El importe pagado equivale a 100 EUR * 1.114700 USD/EUR = 111.47 USD.
    :param impuestos_dr: Nodo condicional para registrar los impuestos aplicables conforme al monto del pago recibido, expresados a la moneda del documento relacionado.
    """

    def __init__(
            self,
            id_documento: str,
            moneda_dr: str,
            num_parcialidad: int,
            imp_saldo_ant: Decimal | int,
            imp_pagado: Decimal | int,
            objeto_imp_dr: str,
            serie: str = None,
            folio: str = None,
            equivalencia_dr: Decimal | int = None,
            impuestos_dr: ImpuestosDR | dict = None,
    ):
        super().__init__({
            'IdDocumento': id_documento,
            'MonedaDR': moneda_dr,
            'NumParcialidad': num_parcialidad,
            'ImpSaldoAnt': imp_saldo_ant,
            'ImpPagado': imp_pagado,
            'ObjetoImpDR': objeto_imp_dr,
            'Serie': serie,
            'Folio': folio,
            'EquivalenciaDR': equivalencia_dr,
            'ImpuestosDR': impuestos_dr,
        })
        

class Pago(ScalarMap):
    """
    Elemento requerido para incorporar la información de la recepción de pagos.

    :param fecha_pago: Atributo requerido para expresar la fecha y hora en la que el beneficiario recibe el pago. Se expresa en la forma aaaa-mm-ddThh:mm:ss, de acuerdo con la especificación ISO 8601.En caso de no contar con la hora se debe registrar 12:00:00.
    :param forma_de_pago_p: Atributo requerido para expresar la clave de la forma en que se realiza el pago.
    :param moneda_p: Atributo requerido para identificar la clave de la moneda utilizada para realizar el pago conforme a la especificación ISO 4217. Cuando se usa moneda nacional se registra MXN. El atributo Pagos:Pago:Monto debe ser expresado en la moneda registrada en este atributo.
    :param monto: Atributo requerido para expresar el importe del pago.
    :param docto_relacionado: Nodo requerido para expresar la lista de documentos relacionados con los pagos. Por cada documento que se relacione se debe generar un nodo DoctoRelacionado.
    :param tipo_cambio_p: Atributo condicional para expresar el tipo de cambio de la moneda a la fecha en que se realizó el pago. El valor debe reflejar el número de pesos mexicanos que equivalen a una unidad de la divisa señalada en el atributo MonedaP. Es requerido cuando el atributo MonedaP es diferente a MXN.
    :param num_operacion: Atributo condicional para expresar el número de cheque, número de autorización, número de referencia, clave de rastreo en caso de ser SPEI, línea de captura o algún número de referencia análogo que identifique la operación que ampara el pago efectuado.
    :param rfc_emisor_cta_ord: Atributo condicional para expresar la clave RFC de la entidad emisora de la cuenta origen, es decir, la operadora, el banco, la institución financiera, emisor de monedero electrónico, etc., en caso de ser extranjero colocar XEXX010101000, considerar las reglas de obligatoriedad publicadas en la página del SAT para éste atributo de acuerdo con el catálogo catCFDI:c_FormaPago.
    :param nom_banco_ord_ext: Atributo condicional para expresar el nombre del banco ordenante, es requerido en caso de ser extranjero. Considerar las reglas de obligatoriedad publicadas en la página del SAT para éste atributo de acuerdo con el catálogo catCFDI:c_FormaPago.
    :param cta_ordenante: Atributo condicional para incorporar el número de la cuenta con la que se realizó el pago. Considerar las reglas de obligatoriedad publicadas en la página del SAT para éste atributo de acuerdo con el catálogo catCFDI:c_FormaPago.
    :param rfc_emisor_cta_ben: Atributo condicional para expresar la clave RFC de la entidad operadora de la cuenta destino, es decir, la operadora, el banco, la institución financiera, emisor de monedero electrónico, etc. Considerar las reglas de obligatoriedad publicadas en la página del SAT para éste atributo de acuerdo con el catálogo catCFDI:c_FormaPago.
    :param cta_beneficiario: Atributo condicional para incorporar el número de cuenta en donde se recibió el pago. Considerar las reglas de obligatoriedad publicadas en la página del SAT para éste atributo de acuerdo con el catálogo catCFDI:c_FormaPago.
    :param tipo_cad_pago: Atributo condicional para identificar la clave del tipo de cadena de pago que genera la entidad receptora del pago. Considerar las reglas de obligatoriedad publicadas en la página del SAT para éste atributo de acuerdo con el catálogo catCFDI:c_FormaPago.
    :param cert_pago: Atributo condicional que sirve para incorporar el certificado que ampara al pago, como una cadena de texto en formato base 64. Es requerido en caso de que el atributo TipoCadPago contenga información.
    :param cad_pago: Atributo condicional para expresar la cadena original del comprobante de pago generado por la entidad emisora de la cuenta beneficiaria. Es requerido en caso de que el atributo TipoCadPago contenga información.
    :param sello_pago: Atributo condicional para integrar el sello digital que se asocie al pago. La entidad que emite el comprobante de pago, ingresa una cadena original y el sello digital en una sección de dicho comprobante, este sello digital es el que se debe registrar en este atributo. Debe ser expresado como una cadena de texto en formato base 64. Es requerido en caso de que el atributo TipoCadPago contenga información.
    :param impuestos_p: Nodo condicional para registrar el resumen de los impuestos aplicables conforme al monto del pago recibido, expresados a la moneda de pago.
    """

    def __init__(
            self,
            fecha_pago: datetime,
            forma_de_pago_p: str,
            moneda_p: str,
            docto_relacionado: Sequence[DoctoRelacionado | dict],
            monto: Decimal | int = None,
            tipo_cambio_p: Decimal | int = None,
            num_operacion: str = None,
            rfc_emisor_cta_ord: str = None,
            nom_banco_ord_ext: str = None,
            cta_ordenante: str = None,
            rfc_emisor_cta_ben: str = None,
            cta_beneficiario: str = None,
            tipo_cad_pago: str = None,
            cert_pago: str = None,
            cad_pago: str = None,
            sello_pago: str = None,
    ):
        super().__init__({
            'FechaPago': fecha_pago,
            'FormaDePagoP': forma_de_pago_p,
            'MonedaP': moneda_p,
            'Monto': monto,
            'TipoCambioP': tipo_cambio_p,
            'NumOperacion': num_operacion,
            'RfcEmisorCtaOrd': rfc_emisor_cta_ord,
            'NomBancoOrdExt': nom_banco_ord_ext,
            'CtaOrdenante': cta_ordenante,
            'RfcEmisorCtaBen': rfc_emisor_cta_ben,
            'CtaBeneficiario': cta_beneficiario,
            'TipoCadPago': tipo_cad_pago,
            'CertPago': cert_pago,
            'CadPago': cad_pago,
            'SelloPago': sello_pago,
            'DoctoRelacionado': docto_relacionado,
        })


# MAIN #
class Pagos(CFDI):
    """
    Complemento para el Comprobante Fiscal Digital por Internet (CFDI) para registrar información sobre la recepción de pagos. El emisor de este complemento para recepción de pagos debe ser quien las leyes le obligue a expedir comprobantes por los actos o actividades que realicen, por los ingresos que se perciban o por las retenciones de contribuciones que efectúen.

    :param pago: Elemento requerido para incorporar la información de la recepción de pagos.
    :return: objeto CFDI
    """

    tag = '{http://www.sat.gob.mx/Pagos20}Pagos'
    version = '2.0'

    def __init__(
            self,
            pago: Sequence[Pago | dict],
    ):
        for p in iterate(pago):
            docto_relacionado = p['DoctoRelacionado']
            for d in iterate(docto_relacionado):
                d['ImpSaldoInsoluto'] = d['ImpSaldoAnt'] - d['ImpPagado']

            p['Monto'] = sum(c["ImpPagado"] / (c.get('EquivalenciaDR') or 1) for c in iterate(docto_relacionado))
            p['ImpuestosP'] = make_impuestos_p(docto_relacionado)

        totales = make_pago_totales(pago)

        super().__init__({
            'Version': self.version,
            'Totales': totales,
            'Pago': pago,
        })



import base64
import hashlib

from lxml import etree

from .ds import SignatureValueType, KeyInfoType, CanonicalizationMethodType, SignatureMethodType, ReferenceType, TransformsType, TransformType, \
    DigestMethodType, X509DataType, X509IssuerSerialType, SignedInfo, Signature
from .. import Signer
from ... import XElement


def _tobytes(element: etree.Element, exclusive=True) -> bytes:
    return etree.tostring(element, method='c14n', exclusive=exclusive)


def _digest(element: etree.Element, exclusive=True) -> str:
    element_digest = hashlib.sha1(
        _tobytes(element, exclusive)
    ).digest()
    return base64.b64encode(element_digest).decode()


def signature_c14n_sha1(signer: Signer, element: etree.Element, nsmap=None) -> XElement:
    digest = _digest(element, exclusive=False)

    signed_info = SignedInfo(
        canonicalization_method=CanonicalizationMethodType(
            algorithm="http://www.w3.org/TR/2001/REC-xml-c14n-20010315",
        ),
        signature_method=SignatureMethodType(
            algorithm="http://www.w3.org/2000/09/xmldsig#rsa-sha1"
        ),
        reference=ReferenceType(
            uri="",
            transforms=TransformsType(
                transform=TransformType(
                    algorithm="http://www.w3.org/2000/09/xmldsig#enveloped-signature"
                )
            ),
            digest_method=DigestMethodType(
                algorithm="http://www.w3.org/2000/09/xmldsig#sha1"
            ),
            digest_value=digest
        )
    )
    signed_info["_nsmap"] = nsmap

    signature = Signature(
        signed_info=signed_info,
        signature_value=SignatureValueType(
            signer.sign_sha1(
                _tobytes(signed_info.to_xml(), exclusive=False)
            )
        ),
        key_info=KeyInfoType(
            x509data=X509DataType(
                x509issuer_serial=X509IssuerSerialType(
                    x509issuer_name=signer.issuer(),
                    x509serial_number=signer.serial_number
                ),
                x509certificate=signer.certificate_base64()
            )
        )
    )
    signature["_nsmap"] = nsmap
    return signature



"""tfd http://www.sat.gob.mx/TimbreFiscalDigital"""
from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import *

from ... import Signer
from ... import CFDI, XElement, ScalarMap
from ...transform import MEXICO_TZ


class TimbreFiscalDigital(CFDI):
    """
    Complemento requerido para el Timbrado Fiscal Digital que da valides a un Comprobante Fiscal Digital.

    :param uuid: Atributo requerido para expresar los 36 caracteres del UUID de la transacción de timbrado
    :param fecha_timbrado: Atributo requerido para expresar la fecha y hora de la generación del timbre
    :param sello_cfd: Atributo requerido para contener el sello digital del comprobante fiscal, que será timbrado. El sello deberá ser expresado cómo una cadena de texto en formato Base 64.
    """

    tag = '{http://www.sat.gob.mx/TimbreFiscalDigital}TimbreFiscalDigital'
    version = '1.0'
    
    def __init__(
            self,
            proveedor: Signer,
            uuid: str,
            fecha_timbrado: datetime,
            sello_cfd: str
    ):
        super().__init__({
            'Version': self.version,
            'UUID': uuid,
            'FechaTimbrado': fecha_timbrado or datetime.now(tz=MEXICO_TZ).replace(tzinfo=None),
            'RfcProvCertif': proveedor.rfc_pac,
            'SelloCFD': sello_cfd,
            'NoCertificadoSAT': proveedor.certificate_number,
            'SelloSAT': '',
        })

        self['SelloSAT'] = proveedor.sign_sha1(
            self.cadena_original().encode()
        )



"""pagoenespecie http://www.sat.gob.mx/pagoenespecie"""
from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ... import CFDI, XElement, ScalarMap


class PagoEnEspecie(CFDI):
    """
    Complemento para la expedición de comprobantes fiscales por la donación en la facilidad fiscal de Pago en Especie
    
    :param cve_pic: Clave de inscripción al Padrón de Instituciones Culturales adheridas al Programa de Pago en Especie
    :param folio_sol_don: Número de folio de la solicitud de donación
    :param pza_art_nombre: Nombre de la pieza de arte
    :param pza_art_tecn: Técnica de producción de la pieza de arte
    :param pza_art_aprod: Año de producción de la pieza de arte
    :param pza_art_dim: Dimensiones de la pieza de arte
    """
    
    tag = '{http://www.sat.gob.mx/pagoenespecie}PagoEnEspecie'
    version = '1.0'
    
    def __init__(
            self,
            cve_pic: str,
            folio_sol_don: str,
            pza_art_nombre: str,
            pza_art_tecn: str,
            pza_art_aprod: str,
            pza_art_dim: str,
    ): 
        super().__init__({
            'Version': self.version,
            'CvePIC': cve_pic,
            'FolioSolDon': folio_sol_don,
            'PzaArtNombre': pza_art_nombre,
            'PzaArtTecn': pza_art_tecn,
            'PzaArtAProd': pza_art_aprod,
            'PzaArtDim': pza_art_dim,
        })
        




import os
from binascii import unhexlify
from datetime import datetime, date
from io import BytesIO
from unittest import mock
from zipfile import ZipInfo

from satcfdi.diot.code import Periodo, TipoOperacion, TipoTercero, Pais
from satcfdi.diot import DatosIdentificacion, DatosComplementaria, ProveedorTercero, DIOT
from tests.utils import get_signer, verify_result

current_dir = os.path.dirname(__file__)
module = 'satcfdi'


def test_encrypt_signature():
    signer = get_signer('cacx7605101p8')
    key = b'\x9e@@hL\xcd\xf2\x9eR\xe6JC\xc8\xba\x8a\xf1C\xec\rE*\x16 4'

    # from DIOT project try 1
    sig = "1a7f1f4d8ef5091a4dc797a80327756110878b08555d84761475e444ad4effe855b44aafb7c47a3627d21f34d40a59fe103d53eda40721d8196d745ce39a6bce97e064a3725840dcf340f7c9ad84f8ab891198f25bf81802f0e8363d4ce02c78268ce75bda284d0771bde7cdf6b6c76adbc607823d54edfa8f9e5385ee8ffa406e14811738781678faff32d0d6b9c843e38ba7a29b8b2414930a200d0512ee69ca693ba341b9462476fbaa085d56178915c03d9bc61ae37d8609fe75c87659e687bd0ca56bcf5a52e2b74fbb67bcd1761ee8360cddb82470a2106a30a9299a7415ccc12a5f37d5b5732e84fd21aba7b3caba49eb913925c9869ca643df1dba8c"
    sig = unhexlify(sig)
    res = signer.decrypt(sig)
    assert res == key

    # from DIOT project try 2
    sig = "768cce9ca1955dbe4b9e768413e37d23f0b1c3e2d3d313278053f2c66aae14d7ecaacdf4a942dff564489156dfb805a3713b7ad48f230762acb493840a1236e445844ba41ca299d715a63dca227f86679322a4864aa30a28b9f43200c83748816900e9b06f224d3ba3bae284ec9916c752da12f0bb189ef1eb64f71bae6ced280f3009798002dcae9129c5105efef1179d93a434c8ebd6144a0f5721deea7c864ed273b8410e2dbdc1d0e684f49c37d58a3311a673ff347f59115ecdf6ba5d05b5e2d2caa45f1e0d2a842057467ed8069de8f79e6abe449c7f1a3222d873d859ad6e6f9608342e1dc709c8b50cb67a07016c75230495938b15e174e0277ac9f1"
    sig = unhexlify(sig)
    res = signer.decrypt(sig)
    assert res == key

    enc_key = signer.encrypt(data=key)
    dec_key = signer.decrypt(data=enc_key)

    assert dec_key == key


def test_create_filename():
    rfc = 'XIQB891116QE4'
    with mock.patch(f'{module}.diot.datetime') as m:
        m.now = mock.Mock(return_value=datetime(2022, 11, 8, 11, 40))
        diot = DIOT(
            datos_identificacion=DatosIdentificacion(
                rfc=rfc,
                ejercicio=2022,
            ),
            periodo=Periodo.OCTUBRE,
            proveedores=[]
        )
        assert diot._tmp_filename() == 'XIQB891116QE40DOTAAM1MCMB811401'

    with mock.patch(f'{module}.diot.datetime') as m:
        p = ProveedorTercero(
            tipo_tercero=TipoTercero.PROVEEDOR_GLOBAL,
            tipo_operacion=TipoOperacion.OTROS
        )
        m.now = mock.Mock(return_value=datetime(2022, 11, 8, 13, 20))
        diot = DIOT(
            datos_identificacion=DatosIdentificacion(
                rfc=rfc,
                ejercicio=2022,
            ),
            periodo=Periodo.OCTUBRE,
            proveedores=[p for _ in range(40001)]
        )
        assert diot._tmp_filename() == 'XIQB891116QE40DOTAAM1MCMB813202'

    rfc = "OÑO120726RX3"
    with mock.patch(f'{module}.diot.datetime') as m:
        m.now = mock.Mock(return_value=datetime(2022, 11, 8, 17, 18))
        diot = DIOT(
            datos_identificacion=DatosIdentificacion(
                rfc=rfc,
                ejercicio=2022,
            ),
            periodo=Periodo.OCTUBRE,
            proveedores=[]
        )
        assert diot._tmp_filename() == '_OÑO120726RX30DOTAAM1MCMB817181'


xiqb891116qe4_issuer = DatosIdentificacion(
    rfc='xiqb891116qe4',
    ejercicio=2022,
    curp='xiqb891116mgrmzr05',
    nombre='BERENICE',
    apellido_paterno='XIMO',
    apellido_materno='QUEZADA'
)


def test_create_declaracion_diot():
    text_file = "XIQB891116QE40DOTAAM1MCMB811401"
    diot = DIOT(
        datos_identificacion=xiqb891116qe4_issuer,
        periodo=Periodo.OCTUBRE
    )
    # Verify TXT
    original_txt_data = open(os.path.join(current_dir, 'diot', f'{text_file}.txt'), 'rb').read()
    with BytesIO() as b:
        diot.plain_write(target=b)
        assert original_txt_data == b.getvalue()

    # Verify FILENAME
    with mock.patch(f'{module}.diot.datetime') as m:
        m.now = mock.Mock(return_value=datetime(2022, 11, 8, 11, 40))
        tmp_filename = diot._tmp_filename()
        assert tmp_filename == text_file

    # Verify ZIP
    def zi(filename):
        return ZipInfo(
            filename=filename,
            date_time=(2022, 11, 8, 11, 41, 8)  # time.localtime(time.time())[:6]
        )

    with mock.patch(f'{module}.zip.ZipInfo', zi) as m:
        my_zip = diot._zip_bytes(tmp_filename)

        original_zip_data = open(os.path.join(current_dir, 'diot', f'{text_file}.zpi'), 'rb').read()
        assert my_zip == original_zip_data

        # Verify DEC
        def encrypt(_, data):
            return b'HB\x9cWQ\xbf\x975i\xfd\xcbH\x1c\xacD\x00pp\'%\xf3"\xa7\x802O\xcagG\x8c6\x07\xa7n9\xe6\x10*P\xd1\x9d\xf8,\x7f\x84T%\x9eR\xea\xab\x00\xeeX7\x00\x87\xfe\xb5\xf7\x9d\xc4\xa2\xbcu\xa4\x13SX\xee\xe6\xfeSv\x16\xcc\xb2y\xbd4nD\xb0\xbd\x10^y\x9f/PZ\x9b\xd1\xc0\x98\x84\xb6\x90\xc1"\xe0\xb9\x83y\x92\x85\x832\x98C\xd0\xe8\x81P\x0b\xde\x14\xaa\x19_\x0c\xf0\xf2\xec\xdc\x9f\x06i'

        def rand(dig):
            if dig == 8:
                return b'p*A\xe4\xcb\x88\xa3\x8f'
            if dig == 24:
                return b'\x9e@@hL\xcd\xf2\x9eR\xe6JC\xc8\xba\x8a\xf1C\xec\rE*\x16 4'

        with mock.patch(f'{module}.Certificate.encrypt', encrypt), mock.patch('os.urandom', rand):
            dec_data = diot._encrypted_bytes(tmp_filename)
            assert dec_data == open(os.path.join(current_dir, 'diot', f'{text_file}.dec'), 'rb').read()

            with mock.patch(f'{module}.diot.datetime') as m:
                m.now = mock.Mock(return_value=datetime(2022, 11, 8, 11, 40))

                res_filename = diot.generate_package(
                    os.path.join(current_dir, 'test_diot')
                )
                assert dec_data == open(os.path.join(current_dir, 'test_diot', f'{text_file}.dec'), 'rb').read()
                assert os.path.basename(res_filename) == f'{text_file}.dec'


def test_create_declaracion_diot_complementary():
    diot_file = "XIQB891116QE40DOTAAM1MCMB816201"
    diot = DIOT(
        datos_identificacion=xiqb891116qe4_issuer,
        periodo=Periodo.MARZO_ABRIL,
        complementaria=DatosComplementaria(
            folio_anterior="6565",
            fecha_presentacion_anterior=date(2022, 11, 7)
        )
    )
    reference_data = open(os.path.join(current_dir, 'diot', f'{diot_file}.txt'), 'rb').read()
    assert reference_data == diot.plain_bytes()


def test_create_declaracion_diot_complementary2():
    diot_file = "XIQB891116QE40DOTAAM1MCMB901581"
    diot = DIOT(
        datos_identificacion=xiqb891116qe4_issuer,
        periodo=Periodo.ENERO,
        complementaria=DatosComplementaria(
            folio_anterior="4567",
            fecha_presentacion_anterior=date(2022, 2, 13)
        )
    )
    reference_data = open(os.path.join(current_dir, 'diot', f'{diot_file}.txt'), 'rb').read()
    assert reference_data == diot.plain_bytes()


def test_create_declaracion_diot_trimestral():
    diot_file = "XIQB891116QE40DOTAAM1MCMB816531"
    diot = DIOT(
        datos_identificacion=xiqb891116qe4_issuer,
        periodo=Periodo.JULIO_SEPTIEMBRE
    )
    reference_data = open(os.path.join(current_dir, 'diot', f'{diot_file}.txt'), 'rb').read()
    assert reference_data == diot.plain_bytes()


ono120726rx3_issue = DatosIdentificacion(
    rfc="OÑO120726RX3",
    razon_social="ORGANICOS ÑAVEZ OSORIO S.A DE C.V",
    ejercicio=2022,
)


def test_create_declaracion_diot_moral():
    diot_file = "_OÑO120726RX30DOTAAM1MCMB817181"
    diot = DIOT(
        datos_identificacion=ono120726rx3_issue,
        periodo=Periodo.MAYO_AGOSTO
    )
    reference_data = open(os.path.join(current_dir, 'diot', f'{diot_file}.txt'), 'rb').read()
    assert reference_data == diot.plain_bytes()


def test_create_declaracion_diot_moral_da():
    diot_file = "_OÑO120726RX30DOTAAM1MCMB820181"
    diot = DIOT(
        datos_identificacion=ono120726rx3_issue,
        periodo=Periodo.OCTUBRE,
        proveedores=[
            ProveedorTercero(
                tipo_tercero=TipoTercero.PROVEEDOR_NACIONAL,
                tipo_operacion=TipoOperacion.PRESTACION_DE_SERVICIOS_PROFESIONALES,
                rfc="LIÑI920228KS8",
                iva16=52000,
                iva16_na=48521,
                iva_rfn=42000,
                iva_rfn_na=32000,
                iva0=31000,
                iva_exento=21000,
                retenido=11000,
                devoluciones=10000
            ),
            ProveedorTercero(
                tipo_tercero=TipoTercero.PROVEEDOR_NACIONAL,
                tipo_operacion=TipoOperacion.ARRENDAMIENTO_DE_INMUEBLES,
                rfc="IVD920810GU2",
                iva16=987,
                iva16_na=876,
                iva_rfn=765,
                iva_rfn_na=654,
                iva0=324,
                iva_exento=251,
                retenido=125,
                devoluciones=444
            ),
        ]
    )
    reference_data = open(os.path.join(current_dir, 'diot', f'{diot_file}.txt'), 'rb').read()
    assert reference_data == diot.plain_bytes()


def test_create_declaracion_diot_moral_full():
    diot_file = "_OÑO120726RX30DOTAAL1LCMB823021"
    diot = DIOT(
        datos_identificacion=DatosIdentificacion(
            rfc="OÑO120726RX3",
            razon_social="ORGANICOS ÑAVEZ OSORIO S.A DE C.V",
            ejercicio=2021,
        ),
        periodo=Periodo.JULIO_SEPTIEMBRE,
        proveedores=[
            ProveedorTercero(
                tipo_tercero=TipoTercero.PROVEEDOR_EXTRANJERO,
                tipo_operacion=TipoOperacion.OTROS,
                id_fiscal="1254TAXID",
                nombre_extranjero="NOMBREEXTRANJERO",
                pais=Pais.ANTIGUA_Y_BERMUDA,
                nacionalidad="BERMUDO",
                iva16=456,
                iva16_na=752,
                iva_rfn=782,
                iva_rfn_na=456,
                iva_import16=123,
                iva_import16_na=475,
                iva_import_exento=7575,
                iva0=45213,
                iva_exento=1247,
                retenido=235,
                devoluciones=786
            ),
            ProveedorTercero(
                tipo_tercero=TipoTercero.PROVEEDOR_GLOBAL,
                tipo_operacion=TipoOperacion.ARRENDAMIENTO_DE_INMUEBLES,
                iva16=9874,
                iva16_na=8521,
                iva_rfn=7632,
                iva_rfn_na=6541,
                iva_import16=5241,
                iva_import16_na=4123,
                iva_import_exento=3562,
                iva0=2415,
                iva_exento=1235,
                retenido=985,
                devoluciones=874
            ),
            ProveedorTercero(
                tipo_tercero=TipoTercero.PROVEEDOR_NACIONAL,
                tipo_operacion=TipoOperacion.OTROS,
                rfc="L&O950913MSA",
                iva16=96208900,
                iva16_na=85100,
                iva_rfn=74300,
                iva_rfn_na=67600,
                iva0=58900,
                iva_exento=47700,
                retenido=36400,
                devoluciones=24864
            ),
        ]
    )
    reference_data = open(os.path.join(current_dir, 'diot', f'{diot_file}.txt'), 'rb').read()
    assert reference_data == diot.plain_bytes()


def verify_invoice(invoice, path):
    verify = verify_result(data=invoice.html_str(), filename=f"{path}.html")
    assert verify


def test_create_declaracion_diot_moral_full_2():
    diot_file = "_OÑO120726RX30DOTAAL1LCMB901151"
    diot = DIOT(
        datos_identificacion=DatosIdentificacion(
            rfc="OÑO120726RX3",
            razon_social="ORGANICOS ÑAVEZ OSORIO S.A DE C.V",
            ejercicio=2021,
        ),
        periodo=Periodo.JULIO_SEPTIEMBRE,
        proveedores=[
            ProveedorTercero(
                tipo_tercero=TipoTercero.PROVEEDOR_EXTRANJERO,
                tipo_operacion=TipoOperacion.OTROS,
                id_fiscal="1254TAXID",
                nombre_extranjero="NOMBREEXTRANJERO",
                pais=Pais.ANTIGUA_Y_BERMUDA,
                nacionalidad="BERMUDO",
                iva16=456,
                iva16_na=752,
                iva_rfn=782,
                iva_rfn_na=456,
                iva_import16=123,
                iva_import16_na=475,
                iva_import_exento=7575,
                iva0=45213,
                iva_exento=1247,
                retenido=235,
                devoluciones=786
            ),
            ProveedorTercero(
                tipo_tercero=TipoTercero.PROVEEDOR_GLOBAL,
                tipo_operacion=TipoOperacion.ARRENDAMIENTO_DE_INMUEBLES,
                iva16=9874,
                iva16_na=8521,
                iva_rfn=7632,
                iva_rfn_na=6541,
                iva_import16=5241,
                iva_import16_na=4123,
                iva_import_exento=3562,
                iva0=2415,
                iva_exento=1235,
                retenido=985,
                devoluciones=874
            ),
            ProveedorTercero(
                tipo_tercero=TipoTercero.PROVEEDOR_NACIONAL,
                tipo_operacion=TipoOperacion.OTROS,
                rfc="L&O950913MSA",
                iva16=96208900,
                iva16_na=85100,
                iva_rfn=74300,
                iva_rfn_na=67600,
                iva0=58900,
                iva_exento=47700,
                retenido=36400,
                devoluciones=24864
            ),
            ProveedorTercero(
                tipo_tercero=TipoTercero.PROVEEDOR_GLOBAL,
                tipo_operacion=TipoOperacion.PRESTACION_DE_SERVICIOS_PROFESIONALES,
                iva16=77757987856,
            ),
            ProveedorTercero(
                tipo_tercero=TipoTercero.PROVEEDOR_NACIONAL,
                tipo_operacion=TipoOperacion.PRESTACION_DE_SERVICIOS_PROFESIONALES,
                rfc="IXS7607092R5",
                iva16_na=500,
                iva_rfn=0
            )
        ]
    )
    reference_data = open(os.path.join(current_dir, 'diot', f'{diot_file}.txt'), 'rb').read()
    assert reference_data == diot.plain_bytes()

    verify_invoice(diot, 'test_create_declaracion_diot_moral_full_2')


def test_create_declaracion_diot_moral_full_2_comp():
    diot = DIOT(
        datos_identificacion=DatosIdentificacion(
            rfc="OÑO120726RX3",
            razon_social="ORGANICOS ÑAVEZ OSORIO S.A DE C.V",
            ejercicio=2021,
        ),
        periodo=Periodo.JULIO_SEPTIEMBRE,
        complementaria=DatosComplementaria(
            folio_anterior="12313",
            fecha_presentacion_anterior=date(2021, 5, 10)
        ),
        proveedores=[
            ProveedorTercero(
                tipo_tercero=TipoTercero.PROVEEDOR_EXTRANJERO,
                tipo_operacion=TipoOperacion.OTROS,
                id_fiscal="1254TAXID",
                nombre_extranjero="NOMBREEXTRANJERO",
                pais=Pais.ANTIGUA_Y_BERMUDA,
                nacionalidad="BERMUDO",
                iva16=456,
                iva16_na=752,
                iva_rfn=782,
                iva_rfn_na=456,
                iva_import16=123,
                iva_import16_na=475,
                iva_import_exento=7575,
                iva0=45213,
                iva_exento=1247,
                retenido=235,
                devoluciones=786
            ),
            ProveedorTercero(
                tipo_tercero=TipoTercero.PROVEEDOR_GLOBAL,
                tipo_operacion=TipoOperacion.ARRENDAMIENTO_DE_INMUEBLES,
                iva16=9874,
                iva16_na=8521,
                iva_rfn=7632,
                iva_rfn_na=6541,
                iva_import16=5241,
                iva_import16_na=4123,
                iva_import_exento=3562,
                iva0=2415,
                iva_exento=1235,
                retenido=985,
                devoluciones=874
            ),
            ProveedorTercero(
                tipo_tercero=TipoTercero.PROVEEDOR_NACIONAL,
                tipo_operacion=TipoOperacion.OTROS,
                rfc="L&O950913MSA",
                iva16=96208900,
                iva16_na=85100,
                iva_rfn=74300,
                iva_rfn_na=67600,
                iva0=58900,
                iva_exento=47700,
                retenido=36400,
                devoluciones=24864
            ),
            ProveedorTercero(
                tipo_tercero=TipoTercero.PROVEEDOR_GLOBAL,
                tipo_operacion=TipoOperacion.PRESTACION_DE_SERVICIOS_PROFESIONALES,
                iva16=77757987856,
            ),
            ProveedorTercero(
                tipo_tercero=TipoTercero.PROVEEDOR_NACIONAL,
                tipo_operacion=TipoOperacion.PRESTACION_DE_SERVICIOS_PROFESIONALES,
                rfc="IXS7607092R5",
                iva16_na=500,
                iva_rfn=0
            )
        ]
    )

    verify_invoice(diot, 'test_create_declaracion_diot_moral_full_2_comp')



import json
import os
from pprint import pprint
from unittest import mock

import pytest


from satcfdi.csf import retrieve, _parse_response
from tests.utils import verify_result

current_dir = os.path.dirname(__file__)
current_filename = os.path.splitext(os.path.basename(__file__))[0]
test_dir = os.path.join(current_dir, current_filename)
module = 'satcfdi'

generacion = [
    'persona_fisica.html',
]


@pytest.mark.parametrize('file', generacion)
def test_get_constancia(file):
    with open(os.path.join(current_dir, 'csf', file), 'rb') as f:
        data = f.read()

    with mock.patch(f'requests.get') as mk:
        mk.return_value.content = data

        res = retrieve('RFC', 'ID_CIF')

        assert mk.called
        mk.call_args.kwargs["headers"]["User-Agent"] = 'this is a test'

        args = json.dumps(mk.call_args.kwargs, indent=2, default=str, ensure_ascii=False)
        verify = verify_result(data=args, filename=f"{file}.json")
        assert verify

    verify = verify_result(data=json.dumps(res, indent=2, default=str, ensure_ascii=False), filename=f"{file}_res.json")
    assert verify


generacion_invalid = [
    'invalid.html'
]


@pytest.mark.parametrize('file', generacion_invalid)
def test_get_constancia_invalid(file):
    with open(os.path.join(current_dir, 'csf', file), 'rb') as f:
        data = f.read()

    with mock.patch(f'requests.get') as mk:
        mk.return_value.content = data

        with pytest.raises(ValueError) as excinfo:
            res = retrieve('RFC', 'ID_CIF')


@pytest.mark.parametrize('file', generacion)
def test_parse_response(file):
    # html = BeautifulSoup(res.text, 'html.parser')
    with open(os.path.join(current_dir, 'csf', file), 'rb') as f:
        data = f.read()

    res = _parse_response(data)
    pprint(res)




"""implocal http://www.sat.gob.mx/implocal"""
from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ... import CFDI, XElement, ScalarMap


class TrasladosLocales(ScalarMap):
    """
    Nodo opcional para la expresión de los impuestos locales trasladados
    
    :param imp_loc_trasladado: Nombre del impuesto local trasladado
    :param tasa_de_traslado: Porcentaje de traslado del impuesto local
    :param importe: Monto del impuesto local trasladado
    """
    
    def __init__(
            self,
            imp_loc_trasladado: str,
            tasa_de_traslado: Decimal | int,
            importe: Decimal | int,
    ): 
        super().__init__({
            'ImpLocTrasladado': imp_loc_trasladado,
            'TasadeTraslado': tasa_de_traslado,
            'Importe': importe,
        })
        

class RetencionesLocales(ScalarMap):
    """
    Nodo opcional para la expresión de los impuestos locales retenidos
    
    :param imp_loc_retenido: Nombre del impuesto local retenido
    :param tasa_de_retencion: Porcentaje de retención del impuesto local
    :param importe: Monto del impuesto local retenido
    """
    
    def __init__(
            self,
            imp_loc_retenido: str,
            tasa_de_retencion: Decimal | int,
            importe: Decimal | int,
    ): 
        super().__init__({
            'ImpLocRetenido': imp_loc_retenido,
            'TasadeRetencion': tasa_de_retencion,
            'Importe': importe,
        })
        

class ImpuestosLocales(CFDI):
    """
    Complemento al Comprobante Fiscal Digital para Impuestos Locales
    
    :param total_de_retenciones: Atributo requerido para expresar la suma total de Retenciones aplicables
    :param total_de_traslados: Atributo requerido para expresar la suma total de traslados aplicables
    :param retenciones_locales: Nodo opcional para la expresión de los impuestos locales retenidos
    :param traslados_locales: Nodo opcional para la expresión de los impuestos locales trasladados
    """
    
    tag = '{http://www.sat.gob.mx/implocal}ImpuestosLocales'
    version = '1.0'
    
    def __init__(
            self,
            total_de_retenciones: Decimal | int,
            total_de_traslados: Decimal | int,
            retenciones_locales: RetencionesLocales | dict = None,
            traslados_locales: TrasladosLocales | dict = None,
    ): 
        super().__init__({
            'Version': self.version,
            'TotaldeRetenciones': total_de_retenciones,
            'TotaldeTraslados': total_de_traslados,
            'RetencionesLocales': retenciones_locales,
            'TrasladosLocales': traslados_locales,
        })
        




from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ... import CFDI, XElement, ScalarMap


class RSAKeyValueType(ScalarMap):
    def __init__(
            self,
            modulus: str,
            exponent: str,
    ):
        """

        :param modulus:
        :param exponent:
        """

        super().__init__({
            'Modulus': modulus,
            'Exponent': exponent,
        })


class DSAKeyValueType(ScalarMap):
    def __init__(
            self,
            p: str,
            q: str,
            y: str,
            seed: str,
            pgen_counter: str,
            g: str = None,
            j: str = None,
    ):
        """

        :param p:
        :param q:
        :param y:
        :param seed:
        :param pgen_counter:
        :param g:
        :param j:
        """

        super().__init__({
            'P': p,
            'Q': q,
            'Y': y,
            'Seed': seed,
            'PgenCounter': pgen_counter,
            'G': g,
            'J': j,
        })


class SignaturePropertyType(ScalarMap):
    def __init__(
            self,
            target: str,
            id: str = None,
    ):
        """

        :param target:
        :param id:
        """

        super().__init__({
            'Target': target,
            'Id': id,
        })


class SignaturePropertiesType(ScalarMap):
    def __init__(
            self,
            signature_property: SignaturePropertyType | dict = None,
            id: str = None,
    ):
        """

        :param signature_property:
        :param id:
        """

        super().__init__({
            'SignatureProperty': signature_property,
            'Id': id,
        })


class ObjectType(ScalarMap):
    def __init__(
            self,
            id: str = None,
            mime_type: str = None,
            encoding: str = None,
    ):
        """

        :param id:
        :param mime_type:
        :param encoding:
        """

        super().__init__({
            'Id': id,
            'MimeType': mime_type,
            'Encoding': encoding,
        })


class SPKIDataType(ScalarMap):
    def __init__(
            self,
            spkisexp: str,
    ):
        """

        :param spkisexp:
        """

        super().__init__({
            'SPKISexp': spkisexp,
        })


class PGPDataType(ScalarMap):
    def __init__(
            self,
            pgpkey_id: str,
            pgpkey_packet: str = None,
    ):
        """

        :param pgpkey_id:
        :param pgpkey_packet:
        """

        super().__init__({
            'PGPKeyID': pgpkey_id,
            'PGPKeyPacket': pgpkey_packet,
        })


class X509IssuerSerialType(ScalarMap):
    def __init__(
            self,
            x509issuer_name: str,
            x509serial_number: int,
    ):
        """

        :param x509issuer_name:
        :param x509serial_number:
        """

        super().__init__({
            'X509IssuerName': x509issuer_name,
            'X509SerialNumber': x509serial_number,
        })


class X509DataType(ScalarMap):
    def __init__(
            self,
            x509issuer_serial: X509IssuerSerialType | dict = None,
            x509ski: str = None,
            x509subject_name: str = None,
            x509certificate: str = None,
            x509crl: str = None,
    ):
        """

        :param x509issuer_serial:
        :param x509ski:
        :param x509subject_name:
        :param x509certificate:
        :param x509crl:
        """

        super().__init__({
            'X509IssuerSerial': x509issuer_serial,
            'X509SKI': x509ski,
            'X509SubjectName': x509subject_name,
            'X509Certificate': x509certificate,
            'X509CRL': x509crl,
        })


class KeyValueType(ScalarMap):
    def __init__(
            self,
            dsakey_value: DSAKeyValueType | dict = None,
            rsakey_value: RSAKeyValueType | dict = None,
    ):
        """

        :param dsakey_value:
        :param rsakey_value:
        """

        super().__init__({
            'DSAKeyValue': dsakey_value,
            'RSAKeyValue': rsakey_value,
        })


class DigestMethodType(ScalarMap):
    def __init__(
            self,
            algorithm: str,
    ):
        """

        :param algorithm:
        """

        super().__init__({
            'Algorithm': algorithm,
        })


class TransformType(ScalarMap):
    def __init__(
            self,
            algorithm: str,
            xpath: str = None,
    ):
        """

        :param algorithm:
        :param xpath:
        """

        super().__init__({
            'Algorithm': algorithm,
            'XPath': xpath,
        })


class TransformsType(ScalarMap):
    def __init__(
            self,
            transform: TransformType | dict = None,
    ):
        """

        :param transform:
        """

        super().__init__({
            'Transform': transform,
        })


class RetrievalMethodType(ScalarMap):
    def __init__(
            self,
            uri: str = None,
            type: str = None,
            transforms: TransformsType | dict = None,
    ):
        """

        :param uri:
        :param type:
        :param transforms:
        """

        super().__init__({
            'URI': uri,
            'Type': type,
            'Transforms': transforms,
        })


class KeyInfoType(ScalarMap):
    def __init__(
            self,
            id: str = None,
            key_name: str = None,
            key_value: KeyValueType | dict = None,
            retrieval_method: RetrievalMethodType | dict = None,
            x509data: X509DataType | dict = None,
            pgpdata: PGPDataType | dict = None,
            spkid_ata: SPKIDataType | dict = None,
            mgmt_data: str = None,
    ):
        """

        :param id:
        :param key_name:
        :param key_value:
        :param retrieval_method:
        :param x509data:
        :param pgpdata:
        :param spkid_ata:
        :param mgmt_data:
        """

        super().__init__({
            'Id': id,
            'KeyName': key_name,
            'KeyValue': key_value,
            'RetrievalMethod': retrieval_method,
            'X509Data': x509data,
            'PGPData': pgpdata,
            'SPKIData': spkid_ata,
            'MgmtData': mgmt_data,
        })


class ReferenceType(ScalarMap):
    def __init__(
            self,
            digest_method: DigestMethodType | dict = None,
            digest_value: str = None,
            id: str = None,
            uri: str = None,
            type: str = None,
            transforms: TransformsType | dict = None,
    ):
        """

        :param digest_method:
        :param digest_value:
        :param id:
        :param uri:
        :param type:
        :param transforms:
        """

        super().__init__({
            'DigestMethod': digest_method,
            'DigestValue': digest_value,
            'Id': id,
            'URI': uri,
            'Type': type,
            'Transforms': transforms,
        })


class ManifestType(ScalarMap):
    def __init__(
            self,
            reference: ReferenceType | dict = None,
            id: str = None,
    ):
        """

        :param reference:
        :param id:
        """

        super().__init__({
            'Reference': reference,
            'Id': id,
        })


class SignatureMethodType(ScalarMap):
    def __init__(
            self,
            algorithm: str,
            hmacoutput_length: int = None,
    ):
        """

        :param algorithm:
        :param hmacoutput_length:
        """

        super().__init__({
            'Algorithm': algorithm,
            'HMACOutputLength': hmacoutput_length,
        })


class CanonicalizationMethodType(ScalarMap):
    def __init__(
            self,
            algorithm: str,
    ):
        """

        :param algorithm:
        """

        super().__init__({
            'Algorithm': algorithm,
        })


class SignedInfoType(ScalarMap):
    def __init__(
            self,
            canonicalization_method: CanonicalizationMethodType | dict = None,
            signature_method: SignatureMethodType | dict = None,
            reference: ReferenceType | dict = None,
            id: str = None,
    ):
        """

        :param canonicalization_method:
        :param signature_method:
        :param reference:
        :param id:
        """

        super().__init__({
            'CanonicalizationMethod': canonicalization_method,
            'SignatureMethod': signature_method,
            'Reference': reference,
            'Id': id,
        })


class SignatureValueType(ScalarMap):
    def __init__(
            self,
            _text: str,
            id: str = None,
    ):
        """

        :param _text:
        :param id:
        """

        super().__init__({
            '_text': _text,
            'Id': id,
        })


class SignatureType(ScalarMap):
    def __init__(
            self,
            signed_info: SignedInfoType | dict = None,
            signature_value: SignatureValueType | dict = None,
            id: str = None,
            key_info: KeyInfoType | dict = None,
            object: ObjectType | dict = None,
    ):
        """

        :param signed_info:
        :param signature_value:
        :param id:
        :param key_info:
        :param object:
        """

        super().__init__({
            'SignedInfo': signed_info,
            'SignatureValue': signature_value,
            'Id': id,
            'KeyInfo': key_info,
            'Object': object,
        })


class RSAKeyValue(RSAKeyValueType, XElement):
    tag = '{http://www.w3.org/2000/09/xmldsig#}RSAKeyValue'


class DSAKeyValue(DSAKeyValueType, XElement):
    tag = '{http://www.w3.org/2000/09/xmldsig#}DSAKeyValue'


class SignatureProperty(SignaturePropertyType, XElement):
    tag = '{http://www.w3.org/2000/09/xmldsig#}SignatureProperty'


class SignatureProperties(SignaturePropertiesType, XElement):
    tag = '{http://www.w3.org/2000/09/xmldsig#}SignatureProperties'


class Manifest(ManifestType, XElement):
    tag = '{http://www.w3.org/2000/09/xmldsig#}Manifest'


class Object(ObjectType, XElement):
    tag = '{http://www.w3.org/2000/09/xmldsig#}Object'


class SPKIData(SPKIDataType, XElement):
    tag = '{http://www.w3.org/2000/09/xmldsig#}SPKIData'


class PGPData(PGPDataType, XElement):
    tag = '{http://www.w3.org/2000/09/xmldsig#}PGPData'


class X509Data(X509DataType, XElement):
    tag = '{http://www.w3.org/2000/09/xmldsig#}X509Data'


class RetrievalMethod(RetrievalMethodType, XElement):
    tag = '{http://www.w3.org/2000/09/xmldsig#}RetrievalMethod'


class KeyValue(KeyValueType, XElement):
    tag = '{http://www.w3.org/2000/09/xmldsig#}KeyValue'


class KeyInfo(KeyInfoType, XElement):
    tag = '{http://www.w3.org/2000/09/xmldsig#}KeyInfo'


class DigestMethod(DigestMethodType, XElement):
    tag = '{http://www.w3.org/2000/09/xmldsig#}DigestMethod'


class Transform(TransformType, XElement):
    tag = '{http://www.w3.org/2000/09/xmldsig#}Transform'


class Transforms(TransformsType, XElement):
    tag = '{http://www.w3.org/2000/09/xmldsig#}Transforms'


class Reference(ReferenceType, XElement):
    tag = '{http://www.w3.org/2000/09/xmldsig#}Reference'


class SignatureMethod(SignatureMethodType, XElement):
    tag = '{http://www.w3.org/2000/09/xmldsig#}SignatureMethod'


class CanonicalizationMethod(CanonicalizationMethodType, XElement):
    tag = '{http://www.w3.org/2000/09/xmldsig#}CanonicalizationMethod'


class SignedInfo(SignedInfoType, XElement):
    tag = '{http://www.w3.org/2000/09/xmldsig#}SignedInfo'


class SignatureValue(SignatureValueType, XElement):
    tag = '{http://www.w3.org/2000/09/xmldsig#}SignatureValue'


class Signature(SignatureType, XElement):
    tag = '{http://www.w3.org/2000/09/xmldsig#}Signature'



"""detallista http://www.sat.gob.mx/detallista"""
from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ... import CFDI, XElement, ScalarMap


class TotalAllowanceCharge(ScalarMap):
    """
    Nodo opcional que especifica el monto total de cargos o descuentos
    
    :param allowance_or_charge_type: Atributo para especificar los cargos o descuentos consolidados
    :param special_services_type: Nodo opcional que especifica el tipo de descuento o cargo. Este atributo sólo se utiliza si el comprador conoce el tipo de descuento o cargo.
    :param amount: Nodo opcional que especifica el monto total de los cargos o descuentos
    """
    
    def __init__(
            self,
            allowance_or_charge_type: str,
            special_services_type: str = None,
            amount: Decimal | int = None,
    ): 
        super().__init__({
            'AllowanceOrChargeType': allowance_or_charge_type,
            'SpecialServicesType': special_services_type,
            'Amount': amount,
        })
        

class TotalAmount(ScalarMap):
    """
    Nodo opcional que especifica el monto total de las líneas de artículos.
    
    :param amount: Nodo requerido que especifica el monto
    """
    
    def __init__(
            self,
            amount: Decimal | int,
    ): 
        super().__init__({
            'Amount': amount,
        })
        

class TradeItemTaxAmount(ScalarMap):
    """
    Nodo opcional que especifica el importe o porcentaje del descuento
    
    :param tax_percentage: Nodo requerido que especifica el porcentaje del impuesto
    :param tax_amount: Nodo requerido que especifica el monto del impuesto
    """
    
    def __init__(
            self,
            tax_percentage: Decimal | int,
            tax_amount: Decimal | int,
    ): 
        super().__init__({
            'TaxPercentage': tax_percentage,
            'TaxAmount': tax_amount,
        })
        

class TradeItemTaxInformation(ScalarMap):
    """
    Nodo opcional que especifica los impuestos por cada línea de artículo
    
    :param tax_type_description: Nodo requerido que especifica el tipo de arancel, impuesto o cuota.
    :param reference_number: Nodo opcional que especifica el numero de identificación del impuesto. Numero asignado al comprador por una jurisdicción de impuesto
    :param trade_item_tax_amount: Nodo opcional que especifica el importe o porcentaje del descuento
    :param tax_category: Nodo opcional que especifica el identificador de impuesto retenido o transferido
    """
    
    def __init__(
            self,
            tax_type_description: str,
            reference_number: str = None,
            trade_item_tax_amount: TradeItemTaxAmount | dict = None,
            tax_category: str = None,
    ): 
        super().__init__({
            'TaxTypeDescription': tax_type_description,
            'ReferenceNumber': reference_number,
            'TradeItemTaxAmount': trade_item_tax_amount,
            'TaxCategory': tax_category,
        })
        

class RatePerUnit(ScalarMap):
    """
    Nodo opcional que especifica la Tarifa por unidad
    
    :param amount_per_unit: Nodo requerido que especifica el importe monetario del cargo o descuento por unidad
    """
    
    def __init__(
            self,
            amount_per_unit: str,
    ): 
        super().__init__({
            'AmountPerUnit': amount_per_unit,
        })
        

class MonetaryAmountOrPercentage(ScalarMap):
    """
    Nodo requerido que especifica la cantidad monetaria o descuento por linea de articulo
    
    :param percentage_per_unit: Nodo requerido que especifica el porcentaje del cargo o descuento
    :param rate_per_unit: Nodo opcional que especifica la Tarifa por unidad
    """
    
    def __init__(
            self,
            percentage_per_unit: str,
            rate_per_unit: RatePerUnit | dict = None,
    ): 
        super().__init__({
            'PercentagePerUnit': percentage_per_unit,
            'RatePerUnit': rate_per_unit,
        })
        

class AllowanceCharge(ScalarMap):
    """
    Nodo opcional que especifica la información de los cargos o descuentos globales por línea de artículo
    
    :param allowance_charge_type: Atributo para especificar el cargo o descuento
    :param monetary_amount_or_percentage: Nodo requerido que especifica la cantidad monetaria o descuento por linea de articulo
    :param settlement_type: Atributo para especificar la imputación del cargo o descuento
    :param sequence_number: Atributo para especificar la secuencia de cálculo
    :param special_services_type: Nodo opcional que especifica el tipo de descuento o cargo. Este atributo sólo se utiliza si el comprador conoce el tipo de descuento o cargo.
    """
    
    def __init__(
            self,
            allowance_charge_type: str,
            monetary_amount_or_percentage: MonetaryAmountOrPercentage | dict,
            settlement_type: str = None,
            sequence_number: str = None,
            special_services_type: str = None,
    ): 
        super().__init__({
            'AllowanceChargeType': allowance_charge_type,
            'MonetaryAmountOrPercentage': monetary_amount_or_percentage,
            'SettlementType': settlement_type,
            'SequenceNumber': sequence_number,
            'SpecialServicesType': special_services_type,
        })
        

class LotNumber(ScalarMap):
    """
    Nodo requerido que especifica el No. De lote
    
    :param _text: 
    :param production_date: Atributo para especificar la fecha de producción
    """
    
    def __init__(
            self,
            _text: str,
            production_date: date = None,
    ): 
        super().__init__({
            '_text': _text,
            'ProductionDate': production_date,
        })
        

class Transport(ScalarMap):
    """
    Nodo requerido que especifica el pago de transporte de embalaje
    
    :param method_of_payment: Nodo requerido que especifica el metodo de pago
    """
    
    def __init__(
            self,
            method_of_payment: str,
    ): 
        super().__init__({
            'MethodOfPayment': method_of_payment,
        })
        

class Description(ScalarMap):
    """
    Nodo requerido que especifica la descripción del empaquetado
    
    :param _text: 
    :param type: Atributo para especificar el tipo de empaquetado
    """
    
    def __init__(
            self,
            _text: str,
            type: str,
    ): 
        super().__init__({
            '_text': _text,
            'Type': type,
        })
        

class PalletInformation(ScalarMap):
    """
    Nodo opcional que especifica la información de empaquetado
    
    :param pallet_quantity: Nodo requerido que especifica el numero de paquetes
    :param description: Nodo requerido que especifica la descripción del empaquetado
    :param transport: Nodo requerido que especifica el pago de transporte de embalaje
    """
    
    def __init__(
            self,
            pallet_quantity: str,
            description: Description | dict,
            transport: Transport | dict,
    ): 
        super().__init__({
            'PalletQuantity': pallet_quantity,
            'Description': description,
            'Transport': transport,
        })
        

class SerialShippingContainerCode(ScalarMap):
    """
    Nodo requerido que especifica la información de Rangos de identificación de productos
    
    :param _text: 
    :param type: Atributo para especificar el codigo del numero de identidad
    """
    
    def __init__(
            self,
            _text: str,
            type: str,
    ): 
        super().__init__({
            '_text': _text,
            'Type': type,
        })
        

class LogisticUnits(ScalarMap):
    """
    Nodo opcional que especifica la información de identificación logística
    
    :param serial_shipping_container_code: Nodo requerido que especifica la información de Rangos de identificación de productos
    """
    
    def __init__(
            self,
            serial_shipping_container_code: SerialShippingContainerCode | dict,
    ): 
        super().__init__({
            'SerialShippingContainerCode': serial_shipping_container_code,
        })
        

class NameAndAddress(ScalarMap):
    """
    Nodo requerido que especifica la etiqueta padre que indica que se empezará a definir el nombre y dirección de la ubicación donde esta la aduana
    
    :param name: Nodo requerido que especifica el nombre de la Aduana
    """
    
    def __init__(
            self,
            name: str,
    ): 
        super().__init__({
            'Name': name,
        })
        

class AlternatePartyIdentification(ScalarMap):
    """
    Nodo requerido que especifica la identificación del no. pedimento a nivel detalle
    
    :param _text: 
    :param type: Atributo para especificar el codigo de identificación secundaria
    """
    
    def __init__(
            self,
            _text: str,
            type: str,
    ): 
        super().__init__({
            '_text': _text,
            'Type': type,
        })
        

class Customs(ScalarMap):
    """
    Nodo opcional que especifica la ubicación donde se especifica el identificador de la aduana
    
    :param alternate_party_identification: Nodo requerido que especifica la identificación del no. pedimento a nivel detalle
    :param reference_date: Nodo requerido que especifica la fecha del pedimento YYYYMMDD
    :param name_and_address: Nodo requerido que especifica la etiqueta padre que indica que se empezará a definir el nombre y dirección de la ubicación donde esta la aduana
    :param gln: Nodo opcional que especifica el número global de localización (GLN) de la aduana
    """
    
    def __init__(
            self,
            alternate_party_identification: AlternatePartyIdentification | dict,
            reference_date: date,
            name_and_address: NameAndAddress | dict,
            gln: str = None,
    ): 
        super().__init__({
            'AlternatePartyIdentification': alternate_party_identification,
            'ReferenceDate': reference_date,
            'NameAndAddress': name_and_address,
            'Gln': gln,
        })
        

class ReferenceIdentification(ScalarMap):
    """
    Nodo opcional que especifica la referencia adicional de los productos
    
    :param _text: 
    :param type: Atributo para especificar el tipo de referencia
    """
    
    def __init__(
            self,
            _text: str,
            type: str,
    ): 
        super().__init__({
            '_text': _text,
            'Type': type,
        })
        

class AdditionalInformation(ScalarMap):
    """
    Nodo opcional que especifica la información adicional de referencia en el detalle de productos
    
    :param reference_identification: Nodo opcional que especifica la referencia adicional de los productos
    """
    
    def __init__(
            self,
            reference_identification: ReferenceIdentification | dict = None,
    ): 
        super().__init__({
            'ReferenceIdentification': reference_identification,
        })
        

class NetPrice(ScalarMap):
    """
    Nodo opcional que especifica la declaración del precion neto
    
    :param amount: Nodo requerido que especifica el precio neto de cada artículo
    """
    
    def __init__(
            self,
            amount: Decimal | int,
    ): 
        super().__init__({
            'Amount': amount,
        })
        

class GrossPrice(ScalarMap):
    """
    Nodo opcional que especifica la declaración del precio bruto
    
    :param amount: Nodo requerido que especifica el precio Bruto del artículo sin incluir descuento ni cargos
    """
    
    def __init__(
            self,
            amount: Decimal | int,
    ): 
        super().__init__({
            'Amount': amount,
        })
        

class AditionalQuantity(ScalarMap):
    """
    Nodo opcional que especifica la cantidad que se esta declarando como adicional
    
    :param _text: 
    :param quantity_type: Atributo para especificar el codigo de tipo de cantidad adicional declarada
    """
    
    def __init__(
            self,
            _text: Decimal | int,
            quantity_type: str,
    ): 
        super().__init__({
            '_text': _text,
            'QuantityType': quantity_type,
        })
        

class TradeItemDescriptionInformation(ScalarMap):
    """
    Nodo opcional que especifica el inicio de la descripción del artículo
    
    :param long_text: Nodo requerido que especifica la descripción del artículo solicitado
    :param language: Atributo para especificar el codigo del idioma en que esta la descripción del articulo
    """
    
    def __init__(
            self,
            long_text: str,
            language: str = None,
    ): 
        super().__init__({
            'LongText': long_text,
            'Language': language,
        })
        

class AlternateTradeItemIdentification(ScalarMap):
    """
    Nodo opcional que especifica el numero de identificación adicional para el artículo.
    
    :param _text: 
    :param type: Atributo para especificar el tipo de identificación adicional, en caso de no utilizar condigo GTIN
    """
    
    def __init__(
            self,
            _text: str,
            type: str,
    ): 
        super().__init__({
            '_text': _text,
            'Type': type,
        })
        

class GrossAmount(ScalarMap):
    """
    Nodo opcional que especifica el importe bruto
    
    :param amount: Nodo requerido que especifica el importe bruto de la línea de artículo
    """
    
    def __init__(
            self,
            amount: Decimal | int,
    ): 
        super().__init__({
            'Amount': amount,
        })
        

class NetAmount(ScalarMap):
    """
    Nodo requerido que especifica el importe neto
    
    :param amount: Nodo requerido que especifica el importe neto de la línea de artículo
    """
    
    def __init__(
            self,
            amount: Decimal | int,
    ): 
        super().__init__({
            'Amount': amount,
        })
        

class TotalLineAmount(ScalarMap):
    """
    Nodo requerido que especifica los importes monetarios por línea de articulo
    
    :param net_amount: Nodo requerido que especifica el importe neto
    :param gross_amount: Nodo opcional que especifica el importe bruto
    """
    
    def __init__(
            self,
            net_amount: NetAmount | dict,
            gross_amount: GrossAmount | dict = None,
    ): 
        super().__init__({
            'NetAmount': net_amount,
            'GrossAmount': gross_amount,
        })
        

class InvoicedQuantity(ScalarMap):
    """
    Nodo requerido que especifica la cantidad facturada del producto en la línea de articulo actual
    
    :param _text: 
    :param unit_of_measure: Atributo para especificar la unidad de medida, solo si el articulo es de unidad variable
    """
    
    def __init__(
            self,
            _text: Decimal | int,
            unit_of_measure: str,
    ): 
        super().__init__({
            '_text': _text,
            'UnitOfMeasure': unit_of_measure,
        })
        

class TradeItemIdentification(ScalarMap):
    """
    Nodo requerido que especifica la identificación de cada artículo
    
    :param gtin: Nodo requerido que especifica el código EAN del artículo solicitado
    """
    
    def __init__(
            self,
            gtin: str,
    ): 
        super().__init__({
            'Gtin': gtin,
        })
        

class LineItem(ScalarMap):
    """
    Nodo opcional que especifica la linea de detalle de la factura
    
    :param trade_item_identification: Nodo requerido que especifica la identificación de cada artículo
    :param invoiced_quantity: Nodo requerido que especifica la cantidad facturada del producto en la línea de articulo actual
    :param total_line_amount: Nodo requerido que especifica los importes monetarios por línea de articulo
    :param type: Atributo para especificar el tipo de línea de detalle
    :param number: Atributo para especificar el numero secuencial que se asigna a cada línea de detalle
    :param alternate_trade_item_identification: Nodo opcional que especifica el numero de identificación adicional para el artículo.
    :param trade_item_description_information: Nodo opcional que especifica el inicio de la descripción del artículo
    :param aditional_quantity: Nodo opcional que especifica la cantidad que se esta declarando como adicional
    :param gross_price: Nodo opcional que especifica la declaración del precio bruto
    :param net_price: Nodo opcional que especifica la declaración del precion neto
    :param additional_information: Nodo opcional que especifica la información adicional de referencia en el detalle de productos
    :param customs: Nodo opcional que especifica la ubicación donde se especifica el identificador de la aduana
    :param logistic_units: Nodo opcional que especifica la información de identificación logística
    :param pallet_information: Nodo opcional que especifica la información de empaquetado
    :param extended_attributes: Nodo opcional que especifica la información adicional de lote del producto facturado
    :param allowance_charge: Nodo opcional que especifica la información de los cargos o descuentos globales por línea de artículo
    :param trade_item_tax_information: Nodo opcional que especifica los impuestos por cada línea de artículo
    """
    
    def __init__(
            self,
            trade_item_identification: TradeItemIdentification | dict,
            invoiced_quantity: InvoicedQuantity | dict,
            total_line_amount: TotalLineAmount | dict,
            type: str = None,
            number: int = None,
            alternate_trade_item_identification: AlternateTradeItemIdentification | dict | Sequence[AlternateTradeItemIdentification | dict] = None,
            trade_item_description_information: TradeItemDescriptionInformation | dict = None,
            aditional_quantity: AditionalQuantity | dict | Sequence[AditionalQuantity | dict] = None,
            gross_price: GrossPrice | dict = None,
            net_price: NetPrice | dict = None,
            additional_information: AdditionalInformation | dict = None,
            customs: Customs | dict | Sequence[Customs | dict] = None,
            logistic_units: LogisticUnits | dict = None,
            pallet_information: PalletInformation | dict = None,
            extended_attributes: LotNumber | dict | Sequence[LotNumber | dict] = None,
            allowance_charge: AllowanceCharge | dict | Sequence[AllowanceCharge | dict] = None,
            trade_item_tax_information: TradeItemTaxInformation | dict | Sequence[TradeItemTaxInformation | dict] = None,
    ): 
        super().__init__({
            'TradeItemIdentification': trade_item_identification,
            'InvoicedQuantity': invoiced_quantity,
            'TotalLineAmount': total_line_amount,
            'Type': type,
            'Number': number,
            'AlternateTradeItemIdentification': alternate_trade_item_identification,
            'TradeItemDescriptionInformation': trade_item_description_information,
            'AditionalQuantity': aditional_quantity,
            'GrossPrice': gross_price,
            'NetPrice': net_price,
            'AdditionalInformation': additional_information,
            'Customs': customs,
            'LogisticUnits': logistic_units,
            'PalletInformation': pallet_information,
            'ExtendedAttributes': extended_attributes,
            'AllowanceCharge': allowance_charge,
            'TradeItemTaxInformation': trade_item_tax_information,
        })
        

class Rate(ScalarMap):
    """
    Nodo opcional que especifica la tarifa
    
    :param base: Atributo para especificar la base del porcentaje que se aplicara
    :param percentage: Nodo opcional que especifica el porcentaje de descuento que se está aplicando según se ha indicado en el campo anterior a toda la factura
    """
    
    def __init__(
            self,
            base: str,
            percentage: Decimal | int,
    ): 
        super().__init__({
            'Base': base,
            'Percentage': percentage,
        })
        

class ShipmentDetail(ScalarMap):
    """
    Nodo opcional que especifica la información pertinente para el embarque de la mercancía.
    
    """
    
    def __init__(
            self,
    ): 
        super().__init__({
        })
        

class DiscountPayment(ScalarMap):
    """
    Nodo opcional que especifica los descuentos por pago
    
    :param discount_type: Atributo para especificar el calificador de porcentajes
    :param percentage: Nodo requerido que especifica el valor de los porcentajes que serán descontados o cargados si la factura no se paga dentro del término. El porcentaje será en base al valor de la factura
    """
    
    def __init__(
            self,
            discount_type: str,
            percentage: str,
    ): 
        super().__init__({
            'DiscountType': discount_type,
            'Percentage': percentage,
        })
        

class TimePeriodDue(ScalarMap):
    """
    Nodo requerido que especifica el tiempo de pago
    
    :param time_period: Atributo para especificar el tipo de periodo
    :param value: Nodo requerido que especifica el numero de periodos
    """
    
    def __init__(
            self,
            time_period: str,
            value: str,
    ): 
        super().__init__({
            'TimePeriod': time_period,
            'Value': value,
        })
        

class PaymentTimePeriod(ScalarMap):
    """
    Nodo opcional que especifica el periodo de pago de la factura
    
    :param time_period_due: Nodo requerido que especifica el tiempo de pago
    """
    
    def __init__(
            self,
            time_period_due: TimePeriodDue | dict,
    ): 
        super().__init__({
            'TimePeriodDue': time_period_due,
        })
        

class NetPayment(ScalarMap):
    """
    Nodo opcional que especifica las condiciones de pago
    
    :param net_payment_terms_type: Atributo para especificar las condiciones de pago
    :param payment_time_period: Nodo opcional que especifica el periodo de pago de la factura
    """
    
    def __init__(
            self,
            net_payment_terms_type: str,
            payment_time_period: PaymentTimePeriod | dict = None,
    ): 
        super().__init__({
            'NetPaymentTermsType': net_payment_terms_type,
            'PaymentTimePeriod': payment_time_period,
        })
        

class PaymentTerms(ScalarMap):
    """
    Nodo opcional que especifica los términos de pago de la factura
    
    :param payment_terms_event: Atributo para especificar la referencia del tiempo de pago
    :param payment_terms_relation_time: Atributo para especificar el termino de relación del tiempo de pago
    :param net_payment: Nodo opcional que especifica las condiciones de pago
    :param discount_payment: Nodo opcional que especifica los descuentos por pago
    """
    
    def __init__(
            self,
            payment_terms_event: str = None,
            payment_terms_relation_time: str = None,
            net_payment: NetPayment | dict = None,
            discount_payment: DiscountPayment | dict = None,
    ): 
        super().__init__({
            'PaymentTermsEvent': payment_terms_event,
            'PaymentTermsRelationTime': payment_terms_relation_time,
            'NetPayment': net_payment,
            'DiscountPayment': discount_payment,
        })
        

class Currency(ScalarMap):
    """
    Nodo opcional que especifica el tipo de divisa utilizada, para efectos de comprobantes fiscales digitales emitidos UNICAMENTE se podrá utilizar como divisa la moneda nacional (MXN), sin embargo dentro del complemento se podrá detallar en otra de forma informativa. Lo detallado en esta etiqueta deberá coincidir con lo declarado en las etiquetas del SAT considerando el tipo de cambio.
    
    :param currency_isocode: Atributo para especificar el codigo de la moneda utilizada para detallar la relación de facturas.Todas las cantidades posteriores seran expresadas en la divisa detallada en esta etiqueta
    :param currency_function: Nodo requerido que especifica la función de divisa
    :param rate_of_change: Nodo opcional que especifical la tasa de cambio que se aplica a las monedas. La regla general para calcular la tasa de cambio es la siguiente: Divisa de Referencia multiplicada por la Tasa = Divisa Objetivo
    """
    
    def __init__(
            self,
            currency_isocode: str,
            currency_function: str | Sequence[str],
            rate_of_change: Decimal | int = None,
    ): 
        super().__init__({
            'CurrencyISOCode': currency_isocode,
            'CurrencyFunction': currency_function,
            'RateOfChange': rate_of_change,
        })
        

class InvoiceCreator(ScalarMap):
    """
    Nodo opcional que especifica la ubicación donde se especifica el identificador del emisor de la factura si es distinto del identificador del proveedor.
    
    :param gln: Nodo opcional que especifica el número global de localización (GLN)de la ubicación del emisor de la factura
    :param alternate_party_identification: Nodo opcional que especifica la identificación secundaria del emisor de la factura
    :param name_and_address:
    """
    
    def __init__(
            self,
            gln: str = None,
            alternate_party_identification: AlternatePartyIdentification | dict = None,
            name_and_address: NameAndAddress | dict = None,
    ): 
        super().__init__({
            'Gln': gln,
            'AlternatePartyIdentification': alternate_party_identification,
            'NameAndAddress': name_and_address,
        })
        

class ShipTo(ScalarMap):
    """
    Nodo opcional que especifica la ubicación donde debe realizarse la entrega de la mercancía.
    
    :param gln: Nodo opcional que especifica el número global de localización (GLN) de la ubicación a entregar la mercancía
    :param name_and_address: Nodo opcional que especifica el nombre y dirección de la ubicación donde debe realizarse la entrega de mercancía
    """
    
    def __init__(
            self,
            gln: str = None,
            name_and_address: NameAndAddress | dict = None,
    ): 
        super().__init__({
            'Gln': gln,
            'NameAndAddress': name_and_address,
        })
        

class Seller(ScalarMap):
    """
    Nodo opcional que especifica información del vendedor
    
    :param gln: Nodo requerido que especifica el número global de localización (GLN) del vendedor.
    :param alternate_party_identification: Nodo requerido que especifica el código para identificar qué tipo de identificación secundaria se le asignó al proveedor
    """
    
    def __init__(
            self,
            gln: str,
            alternate_party_identification: AlternatePartyIdentification | dict,
    ): 
        super().__init__({
            'Gln': gln,
            'AlternatePartyIdentification': alternate_party_identification,
        })
        

class DeliveryNote(ScalarMap):
    """
    Nodo opcional donde se especifica información de recepción de mercancia.Información emitida por el comprador cuando recibe la mercancía que es facturada
    
    :param reference_identification: Nodo requerido que especifica el numero de folio. Número emitido por el comprador cuando recibe la mercancía que es facturada
    :param reference_date: Nodo opcional que especifica la fecha en que fue asignado el no. de folio de recibo
    """
    
    def __init__(
            self,
            reference_identification: str | Sequence[str],
            reference_date: date = None,
    ): 
        super().__init__({
            'ReferenceIdentification': reference_identification,
            'ReferenceDate': reference_date,
        })
        

class SpecialInstruction(ScalarMap):
    """
    Nodo opcional que especifica que tipo de instrucciones comerciales son enviadas
    
    :param code: Atributo para especificar el codigo del tipo de instrucciones comerciales que son enviadas
    :param text: Nodo requerido que especifica información de texto que aplica a todo el mensaje de la factura. La información estará en función al código del tema de texto
    """
    
    def __init__(
            self,
            code: str,
            text: str | Sequence[str],
    ): 
        super().__init__({
            'Code': code,
            'Text': text,
        })
        

class PersonOrDepartmentName(ScalarMap):
    """
    Etiqueta que especifica el contacto de compras
    
    :param text: Contacto de Nodo requerido que especifica el contacto de compras
    """
    
    def __init__(
            self,
            text: str,
    ): 
        super().__init__({
            'Text': text,
        })
        

class ContactInformation(ScalarMap):
    """
    Nodo requerido que especifica información del contacto de compras
    
    :param person_or_department_name: Etiqueta que especifica el contacto de compras
    """
    
    def __init__(
            self,
            person_or_department_name: PersonOrDepartmentName | dict,
    ): 
        super().__init__({
            'PersonOrDepartmentName': person_or_department_name,
        })
        

class Buyer(ScalarMap):
    """
    Nodo requerido que especifica información del comprador
    
    :param gln: Nodo requerido que especifica el número global de localización (GLN) del comprador.
    :param contact_information: Nodo requerido que especifica información del contacto de compras
    """
    
    def __init__(
            self,
            gln: str,
            contact_information: ContactInformation | dict = None,
    ): 
        super().__init__({
            'Gln': gln,
            'ContactInformation': contact_information,
        })
        

class OrderIdentification(ScalarMap):
    """
    Nodo requerido que especifica información sobre la orden de compra a la que hace referencia la factura
    
    :param reference_identification: Nodo requerido que especifica el número de orden de compra (comprador) a la que hace referencia la factura
    :param reference_date: Nodo opcional que especifica la fecha de la orden de compra(comprador) a la que hace referencia la factura
    """
    
    def __init__(
            self,
            reference_identification: ReferenceIdentification | dict | Sequence[ReferenceIdentification | dict],
            reference_date: date = None,
    ): 
        super().__init__({
            'ReferenceIdentification': reference_identification,
            'ReferenceDate': reference_date,
        })
        

class RequestForPaymentIdentification(ScalarMap):
    """
    Nodo requerido que especifica la transacción a utilizar
    
    :param entity_type: Nodo requerido que especifica el tipo de transacción
    """
    
    def __init__(
            self,
            entity_type: str,
    ): 
        super().__init__({
            'EntityType': entity_type,
        })
        

class Detallista(CFDI):
    """
    Complemento requerido para la emision y recepcion de comprobantes fiscales digitales en el sector Retail
    
    :param document_status: Función del mensaje
    :param request_for_payment_identification: Nodo requerido que especifica la transacción a utilizar
    :param order_identification: Nodo requerido que especifica información sobre la orden de compra a la que hace referencia la factura
    :param additional_information: Nodo requerido que especifica las referencias adicionales a nivel global de la factura
    :param buyer: Nodo requerido que especifica información del comprador
    :param type: Tipo de transacción bajo estandar del sector detallista
    :param content_version: Versión del estandar XML utilizado para la elaboración de la guia del sector detallista
    :param special_instruction: Nodo opcional que especifica que tipo de instrucciones comerciales son enviadas
    :param delivery_note: Nodo opcional donde se especifica información de recepción de mercancia.Información emitida por el comprador cuando recibe la mercancía que es facturada
    :param seller: Nodo opcional que especifica información del vendedor
    :param ship_to: Nodo opcional que especifica la ubicación donde debe realizarse la entrega de la mercancía.
    :param invoice_creator: Nodo opcional que especifica la ubicación donde se especifica el identificador del emisor de la factura si es distinto del identificador del proveedor.
    :param customs: Nodo opcional que especifica la ubicación de la aduana
    :param currency: Nodo opcional que especifica el tipo de divisa utilizada, para efectos de comprobantes fiscales digitales emitidos UNICAMENTE se podrá utilizar como divisa la moneda nacional (MXN), sin embargo dentro del complemento se podrá detallar en otra de forma informativa. Lo detallado en esta etiqueta deberá coincidir con lo declarado en las etiquetas del SAT considerando el tipo de cambio.
    :param payment_terms: Nodo opcional que especifica los términos de pago de la factura
    :param shipment_detail: Nodo opcional que especifica la información pertinente para el embarque de la mercancía.
    :param allowance_charge: Nodo opcional que especifica la información de los cargos o descuentos globales mercantiles por factura
    :param line_item: Nodo opcional que especifica la linea de detalle de la factura
    :param total_amount: Nodo opcional que especifica el monto total de las líneas de artículos.
    :param total_allowance_charge: Nodo opcional que especifica el monto total de cargos o descuentos
    """
    
    tag = '{http://www.sat.gob.mx/detallista}detallista'
    document_structure_version = 'AMC8.1'
    
    def __init__(
            self,
            document_status: str,
            request_for_payment_identification: RequestForPaymentIdentification | dict,
            order_identification: OrderIdentification | dict,
            additional_information: ReferenceIdentification | dict | Sequence[ReferenceIdentification | dict],
            buyer: Buyer | dict,
            type: str = None,
            content_version: str = None,
            special_instruction: SpecialInstruction | dict | Sequence[SpecialInstruction | dict] = None,
            delivery_note: DeliveryNote | dict = None,
            seller: Seller | dict = None,
            ship_to: ShipTo | dict = None,
            invoice_creator: InvoiceCreator | dict = None,
            customs: str | Sequence[str] = None,
            currency: Currency | dict | Sequence[Currency | dict] = None,
            payment_terms: PaymentTerms | dict = None,
            shipment_detail: ShipmentDetail | dict = None,
            allowance_charge: AllowanceCharge | dict | Sequence[AllowanceCharge | dict] = None,
            line_item: LineItem | dict | Sequence[LineItem | dict] = None,
            total_amount: TotalAmount | dict = None,
            total_allowance_charge: TotalAllowanceCharge | dict | Sequence[TotalAllowanceCharge | dict] = None,
    ): 
        super().__init__({
            'DocumentStructureVersion': self.document_structure_version,
            'DocumentStatus': document_status,
            'RequestForPaymentIdentification': request_for_payment_identification,
            'OrderIdentification': order_identification,
            'AdditionalInformation': additional_information,
            'Buyer': buyer,
            'Type': type,
            'ContentVersion': content_version,
            'SpecialInstruction': special_instruction,
            'DeliveryNote': delivery_note,
            'Seller': seller,
            'ShipTo': ship_to,
            'InvoiceCreator': invoice_creator,
            'Customs': customs,
            'Currency': currency,
            'PaymentTerms': payment_terms,
            'ShipmentDetail': shipment_detail,
            'AllowanceCharge': allowance_charge,
            'LineItem': line_item,
            'TotalAmount': total_amount,
            'TotalAllowanceCharge': total_allowance_charge,
        })
        




import os
from datetime import datetime
from functools import cache
import pytz
from lxml import etree

# noinspection PyUnresolvedReferences
from .catalog import TRANSLATIONS, CATALOGS
from .helpers import SchemaCollector
from .objectify import cfdi_objectify
from .pdf_environment import DefaultPDFEnvironment
from .schemas import cfdi_schemas
from .xmlify import cfdi_xmlify
from .xslt import TRANSFORMS
from .. import SchemaValidationError, CFDIError, Certificate
from ..models import CertificateStore

__all__ = [
    'TRANSLATIONS',
    'CATALOGS',
    'SchemaCollector',
    'cfdi_xmlify',
    'cfdi_schemas',
    'cfdi_objectify',
    'PDF_INIT_TEMPLATE',
    'MEXICO_TZ',
    'get_timezone',
    'validate_xsd',
    'xslt_transform',
    'verify_certificate',
    'moneda_decimales'
]

current_dir = os.path.dirname(__file__)
SCHEMA_ROOT = os.path.join(current_dir, "schemas")
PDF_INIT_TEMPLATE = DefaultPDFEnvironment.get_template("_init.html")
SAT_Certificate_Store = None
parser = etree.XMLParser(no_network=True, remove_comments=True, remove_blank_text=True)


def moneda_decimales(moneda):
    return CATALOGS['{http://www.sat.gob.mx/sitio_internet/cfd/catalogos}c_Moneda'][moneda][1]


# America/Chihuahua, America/Mazatlan
HUSO_HORARIOS = {
    "Tiempo del Centro en Frontera": "America/Monterrey",
    "Tiempo del Centro": "America/Mexico_City",
    "Tiempo del Noroeste en Frontera": "America/Matamoros",
    "Tiempo del Pacífico Sonora": "America/Hermosillo",
    "Tiempo del Pacífico en Frontera": "America/Tijuana",
    "Tiempo del Pacífico": "America/La_Paz",
    "Tiempo del Sureste": "America/Cancun",
}


@cache
def _pytz_timezone(tz):
    return pytz.timezone(tz)


def get_timezone(codigo_postal):
    tz = HUSO_HORARIOS[
        CATALOGS['{http://www.sat.gob.mx/sitio_internet/cfd/catalogos}c_CodigoPostal'][codigo_postal][4]
    ]

    return _pytz_timezone(tz)


MEXICO_TZ = _pytz_timezone('America/Mexico_City')


def validate_xsd(xml, base):
    element = etree.fromstring(
        '<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema"/>',
        base_url=os.path.join(SCHEMA_ROOT, "schema.xsd"),
        parser=parser
    )

    for col_s in base:
        etree.SubElement(
            element,
            '{http://www.w3.org/2001/XMLSchema}import',
            attrib={
                "namespace": "",
                "schemaLocation": col_s
            }
        )

    schema = etree.XMLSchema(
        element
    )

    if not schema.validate(xml):
        raise SchemaValidationError(schema.error_log)


@cache
def _get_xslt_transform(path):
    return etree.XSLT(
        etree.parse(
            os.path.join(SCHEMA_ROOT, path)
        )
    )


def xslt_transform(tag, version):
    if trans := TRANSFORMS.get(tag):
        path = trans.format(
            v=version.replace(".", "_")
        )
    else:
        raise CFDIError("XSLT File Transform not found")

    return _get_xslt_transform(path)


def verify_certificate(cert: Certificate, at: datetime) -> bool:
    global SAT_Certificate_Store
    if SAT_Certificate_Store is None:
        SAT_Certificate_Store = CertificateStore.create(os.path.join(current_dir, "CertsProd.zip"))
    return SAT_Certificate_Store.verify_certificate(cert.certificate.to_cryptography(), at=at)



from .exceptions import (
    CFDIError,
    # Schema Validation
    SchemaValidationError,
    # Request
    ResponseError,
    DocumentNotFoundError,
    CFDIInvalidError,
    # Transformation
    NamespaceMismatchError
)
from .models import *
from .utils import iterate
from .xelement import XElement
from .cfdi import CFDI
from .utils import ScalarMap



import imaplib
import smtplib
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from os.path import basename
import email
import email.utils
import logging

logger = logging.getLogger(__name__)


class EmailManager:
    def __init__(self, stmp_host, stmp_port, imap_host, imap_port, user, password):
        self.receiver = EmailReceiver(
            host=imap_host,
            port=imap_port,
            user=user,
            password=password
        )
        self.sender = EmailSender(
            host=stmp_host,
            port=stmp_port,
            user=user,
            password=password
        )


class EmailReceiver:
    def __init__(self, host, port, user, password):
        self.host = host
        self.port = port
        self.user = user
        self.password = password

    def __enter__(self):
        self.server = imaplib.IMAP4_SSL(self.host, self.port)
        self.server.login(self.user, password=self.password)

        return self.server

    def __exit__(self, exc_type, exc, exc_tb):
        self.server.close()

    def get_mail_attachments(self, mailbox):
        with self as mail:
            mail.select(
                mailbox=mailbox,
                readonly=False
            )

            typ, data = mail.search(None, 'UnSeen')  # mail.search(None, 'ALL')
            logger.info("%s %s", typ, data)

            for num in data[0].split():
                typ, data = mail.fetch(num, '(RFC822)')
                raw_email = data[0][1]
                # print('Message %s\n%s\n' % (num, raw_email))

                # converts byte literal to string removing b''
                raw_email_string = raw_email.decode('utf-8')
                email_message = email.message_from_string(raw_email_string)

                xml_data = None
                pdf_data = None

                for part in email_message.walk():
                    # this part comes from the snipped I don't understand yet...
                    if part.get_content_maintype() == 'multipart':
                        continue
                    if part.get('Content-Disposition') is None:
                        continue

                    file_name = part.get_filename()

                    if file_name:
                        if file_name.endswith(".xml"):
                            if xml_data:
                                yield xml_data, pdf_data
                                pdf_data = None
                            xml_data = part.get_payload(decode=True)
                        elif file_name.endswith(".pdf"):
                            if pdf_data:
                                yield xml_data, pdf_data
                                xml_data = None
                            pdf_data = part.get_payload(decode=True)

                if xml_data or pdf_data:
                    yield xml_data, pdf_data

                # mail.store(
                #     message_set=num,
                #     command='+FLAGS',
                #     flags='\\Seen'
                # )


class EmailSender:
    def __init__(self, host, port, user, password):
        self.host = host
        self.port = port
        self.user = user
        self.password = password

    def __enter__(self):
        self.server = smtplib.SMTP(host=self.host, port=self.port)
        self.server.starttls()  # Puts connection to SMTP server in TLS mode

        self.server.login(
            user=self.user,
            password=self.password
        )
        return self

    # exc_type, exc, exc_tb
    def __exit__(self, exc_type, exc, exc_tb):
        self.server.close()

    def send_email(self, subject: str, to_addrs: list, html: str = None, file_attachments=None):
        msg = MIMEMultipart()
        msg['From'] = self.user
        msg['To'] = ", ".join(to_addrs)
        msg['Subject'] = subject
        if html:
            msg.attach(MIMEText(html, _subtype='html'))

        for f in file_attachments or []:
            with open(f, "rb") as fil:
                part = MIMEApplication(
                    fil.read(),
                    Name=basename(f)
                )
            # After the file is closed
            part['Content-Disposition'] = 'attachment; filename="%s"' % basename(f)
            msg.attach(part)

        self.server.send_message(
            msg=msg
        )



import json
import os
from datetime import datetime
from decimal import Decimal
from unittest import mock

from satcfdi.pacs import Environment
from satcfdi.pacs.diverza import Diverza
from satcfdi.create.cfd import cfdi33, cfdi40, pago20
from tests.utils import get_signer, verify_result

current_dir = os.path.dirname(__file__)
module = 'satcfdi'


def test_diverza_validate_rfc():
    diverza = Diverza(
        rfc="AAA010101AAA",
        id="3935",
        token="ABCD1234"
    )

    with mock.patch(f'requests.request') as mk:
        mk.return_value.json = mock.Mock(return_value={'identifiers': [{'rfc': 'AAA010101AAA', 'active': 'Y'}]})
        
        r = diverza.rfc_valid("AAA010101AAA")
        assert r

    with mock.patch(f'requests.request') as mk:
        mk.return_value.json = mock.Mock(return_value={'identifiers': [{'rfc': 'AAA010101AAA', 'active': 'Y'}, {'rfc': 'ABC010101DAS', 'active': 'N'}]})
        
        r = diverza.rfc_valid(["AAA010101AAA", "ABC010101DAS"])

        assert r[0]
        assert not r[1]


def test_diverza_stamp():
    signer = get_signer('xiqb891116qe4')
    emisor = cfdi40.Emisor(
        rfc=signer.rfc,
        nombre=signer.legal_name,
        regimen_fiscal="601"
    )

    diverza = Diverza(
        rfc="AAA010101AAA",
        id="3935",
        token="ABCD1234",
        environment=Environment.TEST
    )

    cfdi = cfdi33.Comprobante(
        emisor=emisor,
        lugar_expedicion="56820",
        fecha=datetime.fromisoformat("2022-09-28T22:40:38"),
        receptor=cfdi33.Receptor(
            rfc='KIJ0906199R1',
            nombre='KIJ, S.A DE C.V.',
            uso_cfdi='G03'
        ),
        metodo_pago='PPD',
        serie="A",
        folio="123456",
        conceptos=[
            cfdi33.Concepto(
                cuenta_predial='1234567890',
                clave_prod_serv='10101702',
                cantidad=Decimal('1.00'),
                clave_unidad='E48',
                descripcion='SERVICIOS DE FACTURACION',
                valor_unitario=Decimal('15390.30'),
                impuestos=cfdi33.Impuestos(
                    traslados=cfdi33.Impuesto.parse('002|Tasa|0.160000'),
                    retenciones=[cfdi33.Impuesto.parse('001|Tasa|0.100000'), cfdi33.Impuesto.parse('002|Tasa|0.106667')],
                )
            )
        ]
    )
    cfdi.sign(signer)

    with mock.patch(f'requests.request') as mk:
        mk.return_value.json = mock.Mock(return_value={
            "uuid": "3d94d0c5-9d4d-4c73-a61b-d083c553b7cc",
            "content": "dGhpcyBpcyBhIHRlc3QgeG1s"
        })
        diverza.stamp(cfdi=cfdi)

        assert mk.called
        mk.call_args.kwargs["headers"]["User-Agent"] = 'this is a test'

        args = json.dumps(mk.call_args.kwargs, indent=2)
        verify = verify_result(data=args, filename="test_diverza_stamp.json")
        assert verify

    with mock.patch(f'requests.request') as mk:
        mk.return_value.json = mock.Mock(return_value={
            "uuid": "3d94d0c5-9d4d-4c73-a61b-d083c553b7cc",
            "content": "dGhpcyBpcyBhIHRlc3QgeG1s"
        })
        diverza.issue(
            cfdi=cfdi,
            # certificate_number=signer.certificate_number()
        )

        assert mk.called
        mk.call_args.kwargs["headers"]["User-Agent"] = 'this is a test'

        args = json.dumps(mk.call_args.kwargs, indent=2)
        verify = verify_result(data=args, filename="test_diverza_issue.json")
        assert verify


def test_diverza_stamp_v40():
    signer = get_signer('xiqb891116qe4')
    emisor = cfdi40.Emisor(
        rfc=signer.rfc,
        nombre=signer.legal_name,
        regimen_fiscal="601"
    )

    diverza = Diverza(
        rfc="AAA010101AAA",
        id="3935",
        token="ABCD1234",
        environment=Environment.TEST
    )

    cfdi = cfdi40.Comprobante(
        emisor=emisor,
        lugar_expedicion="56820",
        fecha=datetime.fromisoformat("2022-09-28T22:40:38"),
        receptor=cfdi40.Receptor(
            rfc='KIJ0906199R1',
            nombre='KIJ, S.A DE C.V.',
            uso_cfdi='G03',
            domicilio_fiscal_receptor="59820",
            regimen_fiscal_receptor="601"
        ),
        metodo_pago='PPD',
        serie="A",
        folio="123456",
        conceptos=[
            cfdi40.Concepto(
                cuenta_predial='1234567890',
                clave_prod_serv='10101702',
                cantidad=Decimal('1.00'),
                clave_unidad='E48',
                descripcion='SERVICIOS DE FACTURACION',
                valor_unitario=Decimal('15390.30'),
                impuestos=cfdi40.Impuestos(
                    traslados=cfdi40.Impuesto.parse('002|Tasa|0.160000'),
                    retenciones=[cfdi40.Impuesto.parse('001|Tasa|0.100000'), cfdi40.Impuesto.parse('002|Tasa|0.106667')],
                )
            )
        ]
    )
    cfdi.sign(signer)

    with mock.patch(f'requests.request') as mk:
        mk.return_value.json = mock.Mock(return_value={
            "uuid": "3d94d0c5-9d4d-4c73-a61b-d083c553b7cc",
            "content": "dGhpcyBpcyBhIHRlc3QgeG1s"
        })
        diverza.stamp(cfdi=cfdi)

        assert mk.called
        mk.call_args.kwargs["headers"]["User-Agent"] = 'this is a test'

        args = json.dumps(mk.call_args.kwargs, indent=2)
        verify = verify_result(data=args, filename="test_diverza_stamp_v40.json")
        assert verify

    with mock.patch(f'requests.request') as mk:
        mk.return_value.json = mock.Mock(return_value={
            "uuid": "3d94d0c5-9d4d-4c73-a61b-d083c553b7cc",
            "content": "dGhpcyBpcyBhIHRlc3QgeG1s"
        })
        diverza.issue(
            cfdi=cfdi,
            # certificate_number=signer.certificate_number()
        )

        assert mk.called
        mk.call_args.kwargs["headers"]["User-Agent"] = 'this is a test'

        args = json.dumps(mk.call_args.kwargs, indent=2)
        verify = verify_result(data=args, filename="test_diverza_issue_v40.json")
        assert verify


def test_diverza_stamp_pago_v40():
    signer = get_signer('xiqb891116qe4')
    emisor = cfdi40.Emisor(
        rfc=signer.rfc,
        nombre=signer.legal_name,
        regimen_fiscal="601"
    )

    diverza = Diverza(
        rfc="AAA010101AAA",
        id="3935",
        token="ABCD1234",
        environment=Environment.TEST
    )

    cfdi = cfdi40.Comprobante.pago(
        emisor=emisor,
        fecha=datetime.fromisoformat("2022-09-28T22:40:38"),
        receptor=cfdi40.Receptor(
            rfc='KIJ0906199R1',
            nombre='KIJ, S.A DE C.V.',
            uso_cfdi='G03',
            domicilio_fiscal_receptor="59820",
            regimen_fiscal_receptor="601"
        ),
        lugar_expedicion="56820",
        complemento_pago=pago20.Pagos(
            pago=pago20.Pago(
                fecha_pago=datetime(2020, 1, 1),
                forma_de_pago_p='03',
                moneda_p='MXN',
                tipo_cambio_p=1,
                docto_relacionado=pago20.DoctoRelacionado(
                    id_documento='d6042dc8-d525-4e78-8d1b-092c878bd518',
                    imp_pagado=Decimal("100.3"),
                    imp_saldo_ant=Decimal("203.45"),
                    num_parcialidad=3,
                    moneda_dr="MXN",
                    objeto_imp_dr="01"
                )
            )
        ),
        serie="A",
        folio="123456"
    )
    cfdi.sign(signer)

    with mock.patch(f'requests.request') as mk:
        mk.return_value.json = mock.Mock(return_value={
            "uuid": "3d94d0c5-9d4d-4c73-a61b-d083c553b7cc",
            "content": "dGhpcyBpcyBhIHRlc3QgeG1s"
        })
        diverza.stamp(cfdi=cfdi)

        assert mk.called
        mk.call_args.kwargs["headers"]["User-Agent"] = 'this is a test'

        args = json.dumps(mk.call_args.kwargs, indent=2)
        verify = verify_result(data=args, filename="test_diverza_stamp_pago_v40.json")
        assert verify

    with mock.patch(f'requests.request') as mk:
        mk.return_value.json = mock.Mock(return_value={
            "uuid": "3d94d0c5-9d4d-4c73-a61b-d083c553b7cc",
            "content": "dGhpcyBpcyBhIHRlc3QgeG1s"
        })
        diverza.issue(
            cfdi=cfdi,
            # certificate_number=signer.certificate_number()
        )

        assert mk.called
        mk.call_args.kwargs["headers"]["User-Agent"] = 'this is a test'

        args = json.dumps(mk.call_args.kwargs, indent=2)
        verify = verify_result(data=args, filename="test_diverza_issue_pago_v40.json")
        assert verify




"""iedu http://www.sat.gob.mx/iedu"""
from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ... import CFDI, XElement, ScalarMap


class InstEducativas(CFDI):
    """
    Complemento concepto para la expedición de comprobantes fiscales por parte de Instituciones Educativas Privadas, para los efectos del artículo primero y cuarto del decreto por el que se otorga un estímulo fiscal a las personas físicas en relación con los pagos por servicios educativos
    
    :param nombre_alumno: Atributo requerido para indicar el nombre del Alumno
    :param curp: Atributo requerido para indicar la CURP del alumno de la institución educativa
    :param nivel_educativo: Atributo requerido para indicar el nivel educativo que cursa el alumno
    :param aut_rvoe: Atributo requerido para especificar la clave del centro de trabajo o el reconocimiento de validez oficial de estudios en los términos de la Ley General de Educación que tenga la institución educativa privada donde se realiza el pago.
    :param rfc_pago: Atributo opcional para indicar el RFC de quien realiza el pago cuando sea diferente a quien recibe el servicio
    """
    
    tag = '{http://www.sat.gob.mx/iedu}instEducativas'
    version = '1.0'
    
    def __init__(
            self,
            nombre_alumno: str,
            curp: str,
            nivel_educativo: str,
            aut_rvoe: str,
            rfc_pago: str = None,
    ): 
        super().__init__({
            'Version': self.version,
            'NombreAlumno': nombre_alumno,
            'CURP': curp,
            'NivelEducativo': nivel_educativo,
            'AutRVOE': aut_rvoe,
            'RfcPago': rfc_pago,
        })
        




"""cfdi http://www.sat.gob.mx/cfd/3"""
from collections.abc import *
from decimal import Decimal

from . import cfdi40, pago10
from ... import CFDI, ScalarMap


class Impuesto(cfdi40.Impuesto):
    pass


class CfdiRelacionados(ScalarMap):
    """
    Nodo opcional para precisar la información de los comprobantes relacionados.

    :param tipo_relacion: Atributo requerido para indicar la clave de la relación que existe entre éste que se esta generando y el o los CFDI previos.
    :param cfdi_relacionado: Nodo requerido para precisar la información de los comprobantes relacionados.
    """

    def __init__(
            self,
            tipo_relacion: str,
            cfdi_relacionado: str | Sequence[str],
    ):
        super().__init__({
            'TipoRelacion': tipo_relacion,
            'CfdiRelacionado': cfdi_relacionado,
        })


class Parte(ScalarMap):
    """
    Nodo opcional para expresar las partes o componentes que integran la totalidad del concepto expresado en el comprobante fiscal digital por Internet.

    :param clave_prod_serv: Atributo requerido para expresar la clave del producto o del servicio amparado por la presente parte. Es requerido y deben utilizar las claves del catálogo de productos y servicios, cuando los conceptos que registren por sus actividades correspondan con dichos conceptos.
    :param cantidad: Atributo requerido para precisar la cantidad de bienes o servicios del tipo particular definido por la presente parte.
    :param descripcion: Atributo requerido para precisar la descripción del bien o servicio cubierto por la presente parte.
    :param no_identificacion: Atributo opcional para expresar el número de serie, número de parte del bien o identificador del producto o del servicio amparado por la presente parte. Opcionalmente se puede utilizar claves del estándar GTIN.
    :param unidad: Atributo opcional para precisar la unidad de medida propia de la operación del emisor, aplicable para la cantidad expresada en la parte. La unidad debe corresponder con la descripción de la parte.
    :param valor_unitario: Atributo opcional para precisar el valor o precio unitario del bien o servicio cubierto por la presente parte. No se permiten valores negativos.
    :param importe: Atributo opcional para precisar el importe total de los bienes o servicios de la presente parte. Debe ser equivalente al resultado de multiplicar la cantidad por el valor unitario expresado en la parte. No se permiten valores negativos.
    :param informacion_aduanera: Nodo opcional para introducir la información aduanera aplicable cuando se trate de ventas de primera mano de mercancías importadas o se trate de operaciones de comercio exterior con bienes o servicios.
    """

    def __init__(
            self,
            clave_prod_serv: str,
            cantidad: Decimal | int,
            descripcion: str,
            no_identificacion: str = None,
            unidad: str = None,
            valor_unitario: Decimal | int = None,
            importe: Decimal | int = None,
            informacion_aduanera: str | Sequence[str] = None,
    ):
        super().__init__({
            'ClaveProdServ': clave_prod_serv,
            'Cantidad': cantidad,
            'Descripcion': descripcion,
            'NoIdentificacion': no_identificacion,
            'Unidad': unidad,
            'ValorUnitario': valor_unitario,
            'Importe': importe,
            'InformacionAduanera': informacion_aduanera,
        })


class Traslado(ScalarMap):
    """
    Nodo requerido para la información detallada de un traslado de impuesto específico.

    :param base: Atributo requerido para señalar la suma de los atributos Base de los conceptos del impuesto trasladado. No se permiten valores negativos.
    :param impuesto: Atributo requerido para señalar la clave del tipo de impuesto trasladado.
    :param tipo_factor: Atributo requerido para señalar la clave del tipo de factor que se aplica a la base del impuesto.
    :param tasa_o_cuota: Atributo condicional para señalar el valor de la tasa o cuota del impuesto que se traslada por los conceptos amparados en el comprobante.
    :param importe: Atributo condicional para señalar la suma del importe del impuesto trasladado, agrupado por impuesto, TipoFactor y TasaOCuota. No se permiten valores negativos.
    """

    def __init__(
            self,
            base: Decimal | int,
            impuesto: str,
            tipo_factor: str,
            tasa_o_cuota: Decimal | int = None,
            importe: Decimal | int = None,
    ):
        super().__init__({
            'Base': base,
            'Impuesto': impuesto,
            'TipoFactor': tipo_factor,
            'TasaOCuota': tasa_o_cuota,
            'Importe': importe,
        })


class Retencion(ScalarMap):
    """
    Nodo requerido para la información detallada de un traslado de impuesto específico.

    :param base: Atributo requerido para señalar la suma de los atributos Base de los conceptos del impuesto trasladado. No se permiten valores negativos.
    :param impuesto: Atributo requerido para señalar la clave del tipo de impuesto retencion.
    :param tipo_factor: Atributo requerido para señalar la clave del tipo de factor que se aplica a la base del impuesto.
    :param tasa_o_cuota: Atributo condicional para señalar el valor de la tasa o cuota del impuesto que se traslada por los conceptos amparados en el comprobante.
    :param importe: Atributo condicional para señalar la suma del importe del impuesto trasladado, agrupado por impuesto, TipoFactor y TasaOCuota. No se permiten valores negativos.
    """

    def __init__(
            self,
            base: Decimal | int,
            impuesto: str,
            tipo_factor: str,
            tasa_o_cuota: Decimal | int = None,
            importe: Decimal | int = None,
    ):
        super().__init__({
            'Base': base,
            'Impuesto': impuesto,
            'TipoFactor': tipo_factor,
            'TasaOCuota': tasa_o_cuota,
            'Importe': importe,
        })


class Impuestos(ScalarMap):
    """
    Nodo condicional para expresar el resumen de los impuestos aplicables.

    :param retenciones: Nodo condicional para capturar los impuestos retenidos aplicables. Es requerido cuando en los conceptos se registre algún impuesto retenido.
    :param traslados: Nodo condicional para capturar los impuestos trasladados aplicables. Es requerido cuando en los conceptos se registre un impuesto trasladado.
    """

    def __init__(
            self,
            retenciones: Retencion | dict | str | Sequence[Retencion | dict | str] = None,
            traslados: Traslado | dict | str | Sequence[Traslado | dict | str] = None,
    ):
        super().__init__({
            'Retenciones': retenciones,
            'Traslados': traslados,
        })


class Concepto(ScalarMap):
    """
    Nodo requerido para registrar la información detallada de un bien o servicio amparado en el comprobante.

    :param clave_prod_serv: Atributo requerido para expresar la clave del producto o del servicio amparado por el presente concepto. Es requerido y deben utilizar las claves del catálogo de productos y servicios, cuando los conceptos que registren por sus actividades correspondan con dichos conceptos.
    :param cantidad: Atributo requerido para precisar la cantidad de bienes o servicios del tipo particular definido por el presente concepto.
    :param clave_unidad: Atributo requerido para precisar la clave de unidad de medida estandarizada aplicable para la cantidad expresada en el concepto. La unidad debe corresponder con la descripción del concepto.
    :param descripcion: Atributo requerido para precisar la descripción del bien o servicio cubierto por el presente concepto.
    :param valor_unitario: Atributo requerido para precisar el valor o precio unitario del bien o servicio cubierto por el presente concepto.
    :param no_identificacion: Atributo opcional para expresar el número de parte, identificador del producto o del servicio, la clave de producto o servicio, SKU o equivalente, propia de la operación del emisor, amparado por el presente concepto. Opcionalmente se puede utilizar claves del estándar GTIN.
    :param unidad: Atributo opcional para precisar la unidad de medida propia de la operación del emisor, aplicable para la cantidad expresada en el concepto. La unidad debe corresponder con la descripción del concepto.
    :param descuento: Atributo opcional para representar el importe de los descuentos aplicables al concepto. No se permiten valores negativos.
    :param informacion_aduanera: Nodo opcional para introducir la información aduanera aplicable cuando se trate de ventas de primera mano de mercancías importadas o se trate de operaciones de comercio exterior con bienes o servicios.
    :param cuenta_predial: Nodo opcional para asentar el número de cuenta predial con el que fue registrado el inmueble, en el sistema catastral de la entidad federativa de que trate, o bien para incorporar los datos de identificación del certificado de participación inmobiliaria no amortizable.
    :param complemento_concepto: Nodo opcional donde se incluyen los nodos complementarios de extensión al concepto definidos por el SAT, de acuerdo con las disposiciones particulares para un sector o actividad específica.
    :param parte: Nodo opcional para expresar las partes o componentes que integran la totalidad del concepto expresado en el comprobante fiscal digital por Internet.
    :param traslados: Traslados a aplicar.
    :param retenciones: Retenciones a aplicar.
    :param _traslados_incluidos: si el valor valor_unitario ya incluye traslados.
    """

    def __init__(
            self,
            clave_prod_serv: str,
            cantidad: Decimal | int,
            clave_unidad: str,
            descripcion: str,
            valor_unitario: Decimal | int,
            no_identificacion: str = None,
            unidad: str = None,
            descuento: Decimal | int = None,
            impuestos: Impuestos | dict = None,
            informacion_aduanera: str | Sequence[str] = None,
            cuenta_predial: str = None,
            complemento_concepto: Sequence[CFDI] = None,
            parte: Sequence[Parte | dict] = None,
            _traslados_incluidos: bool = False
    ):
        super().__init__({
            'ClaveProdServ': clave_prod_serv,
            'Cantidad': cantidad,
            'ClaveUnidad': clave_unidad,
            'Descripcion': descripcion,
            'ValorUnitario': valor_unitario,
            'NoIdentificacion': no_identificacion,
            'Unidad': unidad,
            'Descuento': descuento,
            'Impuestos': impuestos,
            'InformacionAduanera': informacion_aduanera,
            'CuentaPredial': cuenta_predial,
            'ComplementoConcepto': complemento_concepto,
            'Parte': parte,
            '_traslados_incluidos': _traslados_incluidos
        })


class Receptor(ScalarMap):
    """
    Nodo requerido para precisar la información del contribuyente receptor del comprobante.

    :param rfc: Atributo requerido para precisar la Clave del Registro Federal de Contribuyentes correspondiente al contribuyente receptor del comprobante.
    :param uso_cfdi: Atributo requerido para expresar la clave del uso que dará a esta factura el receptor del CFDI.
    :param nombre: Atributo opcional para precisar el nombre, denominación o razón social del contribuyente receptor del comprobante.
    :param residencia_fiscal: Atributo condicional para registrar la clave del país de residencia para efectos fiscales del receptor del comprobante, cuando se trate de un extranjero, y que es conforme con la especificación ISO 3166-1 alpha-3. Es requerido cuando se incluya el complemento de comercio exterior o se registre el atributo NumRegIdTrib.
    :param num_reg_id_trib: Atributo condicional para expresar el número de registro de identidad fiscal del receptor cuando sea residente en el extranjero. Es requerido cuando se incluya el complemento de comercio exterior.
    """

    def __init__(
            self,
            rfc: str,
            uso_cfdi: str,
            nombre: str = None,
            residencia_fiscal: str = None,
            num_reg_id_trib: str = None,
    ):
        super().__init__({
            'Rfc': rfc,
            'UsoCFDI': uso_cfdi,
            'Nombre': nombre,
            'ResidenciaFiscal': residencia_fiscal,
            'NumRegIdTrib': num_reg_id_trib,
        })


class Emisor(ScalarMap):
    """
    Nodo requerido para expresar la información del contribuyente emisor del comprobante.

    :param rfc: Atributo requerido para registrar la Clave del Registro Federal de Contribuyentes correspondiente al contribuyente emisor del comprobante.
    :param regimen_fiscal: Atributo requerido para incorporar la clave del régimen del contribuyente emisor al que aplicará el efecto fiscal de este comprobante.
    :param nombre: Atributo opcional para registrar el nombre, denominación o razón social del contribuyente emisor del comprobante.
    """

    def __init__(
            self,
            rfc: str,
            regimen_fiscal: str,
            nombre: str = None,
    ):
        super().__init__({
            'Rfc': rfc,
            'RegimenFiscal': regimen_fiscal,
            'Nombre': nombre,
        })


class Comprobante(cfdi40.Comprobante):
    tag = '{http://www.sat.gob.mx/cfd/3}Comprobante'
    version = '3.3'
    complemento_pago = pago10.Pagos



import os
import zipfile
from collections.abc import Iterable
from datetime import datetime

from cryptography import x509
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives.asymmetric import padding
from pytz import utc

current_dir = os.path.dirname(__file__)


class CertificateStore:
    def __init__(self, trusted_certificates: Iterable):
        self._trusted_certificates = {
            cert.subject.public_bytes(): cert for cert in trusted_certificates
        }

    def issuer_certificate(self, cert):
        return self._trusted_certificates[cert.issuer.public_bytes()]

    def verify_certificate(self, issued_cert, at: datetime):
        """Verifies issued_cert.

        Args:
            issued_cert: The issued certificate.

        Returns:
            True if certificate was issued by a trusted certificate
            :param issued_cert: issued certificate to validate
            :param at: date at witch to do the validation
        """
        if at.tzinfo:
            at = at.astimezone(tz=utc).replace(tzinfo=None)

        def validate_date(cert):
            if not cert.not_valid_before <= at <= cert.not_valid_after:
                raise ValueError("Date Not Valid")

        try:
            validate_date(issued_cert)
            signing_cert = self.issuer_certificate(issued_cert)
            validate_date(signing_cert)

            signing_cert.public_key().verify(
                signature=issued_cert.signature,
                data=issued_cert.tbs_certificate_bytes,
                padding=padding.PKCS1v15(),
                algorithm=issued_cert.signature_hash_algorithm
            )

            # Assume the parent certificates have been already validated, only checking dates
            while True:
                parent_cert = self.issuer_certificate(signing_cert)
                if parent_cert == signing_cert:
                    return True
                validate_date(parent_cert)
                signing_cert = parent_cert

        except (IndexError, InvalidSignature, ValueError) as ex:
            return False

    @classmethod
    def create(cls, certs_zip):
        """
        Creates a Certificate Store from the certificates in a zip file
        :param certs_zip:
        :return: CertificateStore
        """
        with zipfile.ZipFile(certs_zip, "r") as zf:
            return cls(
                trusted_certificates=
                (x509.load_der_x509_certificate(zf.read(fileinfo)) for fileinfo in zf.infolist())
            )



# ANSI COLORS
COLOR_RESET = 0

COLOR_BOLD = 1
COLOR_FAINT = 2
COLOR_ITALIC = 3
COLOR_UNDERLINE = 4
COLOR_RAPID_BLINK = 5
COLOR_SLOW_BLINK = 6
COLOR_REVERSED = 7
COLOR_NOT_UNDERLINE = 24

COLOR_BLACK = 30
COLOR_RED = 31
COLOR_GREEN = 32
COLOR_YELLOW = 33
COLOR_BLUE = 34
COLOR_MAGENTA = 35
COLOR_CYAN = 36
COLOR_WHITE = 37
COLOR_DEFAULT = 39

COLOR_BRIGHT_BLACK = 90
COLOR_BRIGHT_RED = 91
COLOR_BRIGHT_GREEN = 92
COLOR_BRIGHT_YELLOW = 93
COLOR_BRIGHT_BLUE = 94
COLOR_BRIGHT_MAGENTA = 95
COLOR_BRIGHT_CYAN = 96
COLOR_BRIGHT_WHITE = 97

COLOR_BACKGROUND_BLACK = 40
COLOR_BACKGROUND_RED = 41
COLOR_BACKGROUND_GREEN = 42
COLOR_BACKGROUND_YELLOW = 43
COLOR_BACKGROUND_BLUE = 44
COLOR_BACKGROUND_MAGENTA = 45
COLOR_BACKGROUND_CYAN = 46
COLOR_BACKGROUND_WHITE = 47
COLOR_BACKGROUND_DEFAULT = 49

COLOR_BACKGROUND_BRIGHT_BLACK = 100
COLOR_BACKGROUND_BRIGHT_RED = 101
COLOR_BACKGROUND_BRIGHT_GREEN = 102
COLOR_BACKGROUND_BRIGHT_YELLOW = 103
COLOR_BACKGROUND_BRIGHT_BLUE = 104
COLOR_BACKGROUND_BRIGHT_MAGENTA = 105
COLOR_BACKGROUND_BRIGHT_CYAN = 106
COLOR_BACKGROUND_BRIGHT_WHITE = 107


def ansi_color(*colors):
    color = ";".join(str(c) for c in colors if c is not None)
    return f"\033[{color}m"


def paint(text, *colors):
    return ansi_color(*colors) + str(text) + ansi_color(COLOR_DEFAULT)




import re
from enum import Enum, auto

# Persona Física / Persona Moral
RFC_Regex = "([A-Z&Ñ]{3,4})([0-9]{2})(0[1-9]|1[012])(0[1-9]|[12][0-9]|3[01])([A-Z0-9]{2}[0-9A])"

RFC_Verify_Chars = "0123456789ABCDEFGHIJKLMN&OPQRSTUVWXYZ Ñ"


class RFCType(Enum):
    FISICA = auto()
    MORAL = auto()


class RFC(str):
    def __new__(cls, value: str, entity_type: RFCType = None):
        value = value.upper()
        self = super().__new__(cls, value)

        match = re.fullmatch(RFC_Regex, value)
        if match:
            if len(match.group(1)) == 3:
                self.type = RFCType.MORAL
            else:
                self.type = RFCType.FISICA
        else:
            raise ValueError("RFC Not Valid")

        if entity_type and self.type != entity_type:
            raise ValueError("RFC Not as Expected ", self.type)

        return self

    def is_generic(self):
        return self in [RFC_Generico_Nacional, RFC_Generico_Extranjero]

    def is_valid(self):
        if self.is_generic():
            return True

        tot = 0
        if self.type == RFCType.MORAL:
            tot += 481

        for i, c in enumerate(self[::-1], start=1):
            tot += RFC_Verify_Chars.index(c) * i

        return tot % 11 == 0

    def __repr__(self):
        return '%s(%s)' % (
            self.__class__.__qualname__,
            f'{super().__repr__()}'
        )


RFC_Generico_Nacional = RFC("XAXX010101000")
RFC_Generico_Extranjero = RFC("XEXX010101000")



import io
from collections import namedtuple
from zipfile import ZipFile, ZipInfo

ZipData = namedtuple("ZipFile", "filename data")


def zip_create(target: io.BytesIO, files: list[ZipData]):
    p = target.tell()

    for f in files:
        with ZipFile(target, "w") as myzip:
            zinfo = ZipInfo(
                filename=f.filename
            )
            zinfo.filename = f.filename
            zinfo.compress_type = 8
            zinfo.create_system = 0

            with myzip.open(zinfo, 'w') as stream:
                zinfo.flag_bits = 2056
                zinfo.external_attr = 0
                f.data(stream)

    with target.getbuffer() as view:  # change zip flag bytes
        view[p + 6:p + 8] = b"\x08\x08"
